# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from sklearn.model_selection import train_test_split

from mlxtend.classifier import OneRClassifier
from mlxtend.data import iris_data

X, y = iris_data()


def get_feature_quartiles(X):
    X_discretized = X.copy()
    for col in range(X.shape[1]):
        for q, class_label in zip([1.0, 0.75, 0.5, 0.25], [3, 2, 1, 0]):
            threshold = np.quantile(X[:, col], q=q)
            X_discretized[X[:, col] <= threshold, col] = class_label
    return X_discretized.astype(np.int_)


Xd = get_feature_quartiles(X)
Xd_train, Xd_test, y_train, y_test = train_test_split(Xd, y, random_state=0, stratify=y)


def test_iris_quartiles_resolve_ties_first():
    oner = OneRClassifier()
    oner.fit(Xd_train, y_train)
    assert oner.feature_idx_ == 2
    assert oner.prediction_dict_["total error"] == 16
    assert round(oner.score(Xd_train, y_train), 4) == 0.8571
    assert round(oner.score(Xd_test, y_test), 4) == 0.8421


def test_iris_quartiles_resolve_ties_chi_squared_1():
    oner = OneRClassifier(resolve_ties="chi-squared")
    oner.fit(Xd_train, y_train)
    assert oner.feature_idx_ == 2
    assert oner.prediction_dict_["total error"] == 16
    assert round(oner.score(Xd_train, y_train), 4) == 0.8571
    assert round(oner.score(Xd_test, y_test), 4) == 0.8421
    np.testing.assert_almost_equal(oner.p_value_, 0.0, decimal=7)


def test_iris_quartiles_resolve_ties_chi_squared_2():
    # tests with duplicate column
    oner = OneRClassifier(resolve_ties="chi-squared")

    Xd_traintemp = np.zeros((Xd_train.shape[0], Xd_train.shape[1] + 1))
    Xd_traintemp[:, 0] = Xd_train[:, 2]
    Xd_traintemp[:, 1] = Xd_train[:, 0]
    Xd_traintemp[:, 2] = Xd_train[:, 1]
    Xd_traintemp[:, 3] = Xd_train[:, 2]
    Xd_traintemp[:, 4] = Xd_train[:, 3]

    oner.fit(Xd_traintemp, y_train)
    assert oner.feature_idx_ == 0
    assert oner.prediction_dict_["total error"] == 16
    assert round(oner.score(Xd_traintemp, y_train), 4) == 0.8571



import unittest

import numpy as np
import pandas as pd
from test_fpbase import (
    FPTestEdgeCases,
    FPTestErrors,
    FPTestEx1,
    FPTestEx2,
    FPTestEx3All,
    compare_dataframes,
)

from mlxtend.frequent_patterns import fpmax


class TestEdgeCases(unittest.TestCase, FPTestEdgeCases):
    def setUp(self):
        FPTestEdgeCases.setUp(self, fpmax)


class TestErrors(unittest.TestCase, FPTestErrors):
    def setUp(self):
        FPTestErrors.setUp(self, fpmax)


class TestEx1(unittest.TestCase, FPTestEx1):
    def setUp(self, one_ary=None):
        FPTestEx1.setUp(self, fpmax, one_ary=one_ary)

    def test_default(self):
        res_df = fpmax(self.df)
        expect = pd.DataFrame(
            [
                [0.6, frozenset([5, 6])],
                [0.6, frozenset([5, 10])],
                [0.6, frozenset([3, 5, 8])],
            ],
            columns=["support", "itemsets"],
        )

        compare_dataframes(res_df, expect)

    def test_max_len(self):
        res_df1 = fpmax(self.df)
        max_len = np.max(res_df1["itemsets"].apply(len))
        assert max_len == 3

        res_df2 = fpmax(self.df, max_len=2)
        max_len = np.max(res_df2["itemsets"].apply(len))
        assert max_len == 2


class TestEx1BoolInput(TestEx1):
    def setUp(self):
        one_ary = np.array(
            [
                [False, False, False, True, False, True, True, True, True, False, True],
                [False, False, True, True, False, True, False, True, True, False, True],
                [
                    True,
                    False,
                    False,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    True,
                    True,
                ],
                [
                    False,
                    True,
                    False,
                    True,
                    True,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                ],
            ]
        )
        FPTestEx1.setUp(self, fpmax, one_ary=one_ary)


class TestEx2(unittest.TestCase, FPTestEx2):
    def setUp(self):
        FPTestEx2.setUp(self)

    def test_output(self):
        res_df = fpmax(self.df, min_support=0.001, use_colnames=True)
        expect = pd.DataFrame(
            [
                [0.25, frozenset(["a"])],
                [0.25, frozenset(["b"])],
                [0.25, frozenset(["c", "d"])],
                [0.25, frozenset(["e"])],
            ],
            columns=["support", "itemsets"],
        )

        compare_dataframes(res_df, expect)


class TestEx3(unittest.TestCase, FPTestEx3All):
    def setUp(self):
        FPTestEx3All.setUp(self, fpmax)


class TestEx4(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            [[1, 1, 0], [1, 0, 1], [0, 0, 1]], columns=["a", "b", "c"]
        )
        self.fpalgo = fpmax

    def test_output(self):
        res_df = self.fpalgo(self.df, min_support=0.01, use_colnames=True)
        expect = pd.DataFrame(
            [
                [0.3333333333333333, frozenset(["a", "b"])],
                [0.3333333333333333, frozenset(["a", "c"])],
            ],
            columns=["support", "itemsets"],
        )

        compare_dataframes(res_df, expect)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Nonparametric Permutation Test
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


import os

import pandas as pd
import pytest
from sklearn.ensemble import BaggingClassifier, BaggingRegressor
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from mlxtend.data import boston_housing_data, iris_data
from mlxtend.evaluate import bias_variance_decomp


def pandas_input_fail():
    X, y = iris_data()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123, shuffle=True, stratify=y
    )

    X_train = pd.DataFrame(X_train)

    tree = DecisionTreeClassifier(random_state=123)

    with pytest.raises(ValueError):
        avg_expected_loss, avg_bias, avg_var = bias_variance_decomp(
            tree, X_train, y_train, X_test, y_test, loss="0-1_loss", random_seed=123
        )


def test_01_loss_tree():
    X, y = iris_data()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123, shuffle=True, stratify=y
    )

    tree = DecisionTreeClassifier(random_state=123)
    avg_expected_loss, avg_bias, avg_var = bias_variance_decomp(
        tree, X_train, y_train, X_test, y_test, loss="0-1_loss", random_seed=123
    )

    assert round(avg_expected_loss, 3) == 0.062
    assert round(avg_bias, 3) == 0.022
    assert round(avg_var, 3) == 0.040


def test_01_loss_bagging():
    X, y = iris_data()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123, shuffle=True, stratify=y
    )

    tree = DecisionTreeClassifier(random_state=123)
    bag = BaggingClassifier(base_estimator=tree, random_state=123)
    avg_expected_loss, avg_bias, avg_var = bias_variance_decomp(
        bag, X_train, y_train, X_test, y_test, loss="0-1_loss", random_seed=123
    )

    assert round(avg_expected_loss, 3) == 0.048
    assert round(avg_bias, 3) == 0.022
    assert round(avg_var, 3) == 0.026


def test_mse_tree():
    X, y = boston_housing_data()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123, shuffle=True
    )

    tree = DecisionTreeRegressor(random_state=123)
    avg_expected_loss, avg_bias, avg_var = bias_variance_decomp(
        tree, X_train, y_train, X_test, y_test, loss="mse", random_seed=123
    )

    assert round(avg_expected_loss, 3) == 31.536
    assert round(avg_bias, 3) == 14.096
    assert round(avg_var, 3) == 17.440


def test_mse_bagging():
    X, y = boston_housing_data()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123, shuffle=True
    )

    tree = DecisionTreeRegressor(random_state=123)
    bag = BaggingRegressor(base_estimator=tree, n_estimators=10, random_state=123)

    avg_expected_loss, avg_bias, avg_var = bias_variance_decomp(
        bag, X_train, y_train, X_test, y_test, loss="mse", random_seed=123
    )

    assert round(avg_expected_loss, 2) == 20.24, avg_expected_loss
    assert round(avg_bias, 2) == 15.63, avg_bias
    assert round(avg_var, 2) == 4.61, avg_var


if "TRAVIS" in os.environ or os.environ.get("TRAVIS") == "true":
    TRAVIS = True
else:
    TRAVIS = False

if "APPVEYOR" in os.environ or os.environ.get("APPVEYOR") == "true":
    APPVEYOR = True
else:
    APPVEYOR = False


@pytest.mark.skipif(TRAVIS or APPVEYOR, reason="TensorFlow dependency")
def test_keras():
    import tensorflow as tf

    X, y = boston_housing_data()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123, shuffle=True
    )

    model = tf.keras.Sequential(
        [tf.keras.layers.Dense(32, activation=tf.nn.relu), tf.keras.layers.Dense(1)]
    )

    optimizer = tf.keras.optimizers.Adam()
    model.compile(loss="mean_squared_error", optimizer=optimizer)

    model.fit(X_train, y_train, epochs=10)
    model.predict(X_test)



# mlxtend Machine Learning Library Extensions
# Author: Steve Harenberg <harenbergsd@gmail.com>
#
# License: BSD 3 clause

import collections
import math

from ..frequent_patterns import fpcommon as fpc


def fpmax(df, min_support=0.5, use_colnames=False, max_len=None, verbose=0):
    """Get maximal frequent itemsets from a one-hot DataFrame

    Parameters
    -----------
    df : pandas DataFrame
      pandas DataFrame the encoded format. Also supports
      DataFrames with sparse data; for more info, please
      see (https://pandas.pydata.org/pandas-docs/stable/
           user_guide/sparse.html#sparse-data-structures)

      Please note that the old pandas SparseDataFrame format
      is no longer supported in mlxtend >= 0.17.2.

      The allowed values are either 0/1 or True/False.
      For example,

    ```
           Apple  Bananas   Beer  Chicken   Milk   Rice
        0   True    False   True     True  False   True
        1   True    False   True    False  False   True
        2   True    False   True    False  False  False
        3   True     True  False    False  False  False
        4  False    False   True     True   True   True
        5  False    False   True    False   True   True
        6  False    False   True    False   True  False
        7   True     True  False    False  False  False
    ```

    min_support : float (default: 0.5)
      A float between 0 and 1 for minimum support of the itemsets returned.
      The support is computed as the fraction
      transactions_where_item(s)_occur / total_transactions.

    use_colnames : bool (default: False)
      If true, uses the DataFrames' column names in the returned DataFrame
      instead of column indices.

    max_len : int (default: None)
      Given the set of all maximal itemsets,
      return those that are less than `max_len`. If `None` (default) all
      possible itemsets lengths are evaluated.

    verbose : int (default: 0)
      Shows the stages of conditional tree generation.

    Returns
    -----------
    pandas DataFrame with columns ['support', 'itemsets'] of all maximal
      itemsets that are >= `min_support` and < than `max_len`
      (if `max_len` is not None).
      Each itemset in the 'itemsets' column is of type `frozenset`,
      which is a Python built-in type that behaves similarly to
      sets except that it is immutable
      (For more info, see
      https://docs.python.org/3.6/library/stdtypes.html#frozenset).

    Examples
    ----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/fpmax/

    """
    fpc.valid_input_check(df)

    if min_support <= 0.0:
        raise ValueError(
            "`min_support` must be a positive "
            "number within the interval `(0, 1]`. "
            "Got %s." % min_support
        )

    colname_map = None
    if use_colnames:
        colname_map = {idx: item for idx, item in enumerate(df.columns)}

    tree, rank = fpc.setup_fptree(df, min_support)

    minsup = math.ceil(min_support * len(df))  # min support as count
    generator = fpmax_step(tree, minsup, MFITree(rank), colname_map, max_len, verbose)

    return fpc.generate_itemsets(generator, len(df), colname_map)


def fpmax_step(tree, minsup, mfit, colnames, max_len, verbose):
    count = 0
    items = list(tree.nodes.keys())
    largest_set = sorted(tree.cond_items + items, key=mfit.rank.get)
    if len(largest_set) == 0:
        return
    if tree.is_path():
        if not mfit.contains(largest_set):
            count += 1
            largest_set.reverse()
            mfit.cache = largest_set
            mfit.insert_itemset(largest_set)
            if max_len is None or len(largest_set) <= max_len:
                support = tree.root.count
                if len(items) > 0:
                    support = min([tree.nodes[i][0].count for i in items])
                yield support, largest_set

    if verbose:
        tree.print_status(count, colnames)

    if not tree.is_path() and (not max_len or max_len > len(tree.cond_items)):
        # Loop over each item in tree creating another conditional tree
        items.sort(key=tree.rank.get)
        for item in items:
            # Check if the tree will produce a subset already produced
            if mfit.contains(largest_set):
                return
            largest_set.remove(item)
            cond_tree = tree.conditional_tree(item, minsup)
            for support, mfi in fpmax_step(
                cond_tree, minsup, mfit, colnames, max_len, verbose
            ):
                yield support, mfi


class MFITree(object):
    def __init__(self, rank):
        self.root = self.Node(None)
        self.nodes = collections.defaultdict(list)
        self.cache = []
        self.rank = rank

    def insert_itemset(self, itemset, count=1):
        """
        Inserts a list of items into the tree.

        Parameters
        ----------
        itemset : list
            Items that will be inserted into the tree.
        count : int
            The number of occurrences of the itemset.
        """
        if len(itemset) == 0:
            return

        # Follow existing path in tree as long as possible
        index = 0
        node = self.root
        for item in itemset:
            if item in node.children:
                child = node.children[item]
                node = child
                index += 1
            else:
                break

        # Insert any remaining items
        for item in itemset[index:]:
            child_node = self.Node(item, count, node)
            self.nodes[item].append(child_node)
            node = child_node

    def contains(self, itemset):
        """
        Checks if this tree contains itemset in one of its branches.
        The algorithm assumes that  itemset is sorted according to self.rank.
        """
        i = 0
        for item in reversed(self.cache):
            if self.rank[itemset[i]] < self.rank[item]:
                break
            if itemset[i] == item:
                i += 1
            if i == len(itemset):
                return True

        for basenode in self.nodes[itemset[0]]:
            i = 0
            node = basenode
            while node.item is not None:
                if self.rank[itemset[i]] < self.rank[node.item]:
                    break
                if itemset[i] == node.item:
                    i += 1
                if i == len(itemset):
                    return True
                node = node.parent

        return False

    class Node(object):
        def __init__(self, item, count=1, parent=None):
            self.item = item
            self.parent = parent
            self.children = collections.defaultdict(MFITree.Node)

            if parent is not None:
                parent.children[item] = self



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


import numpy as np

from mlxtend.math import vectorspace_dimensionality, vectorspace_orthonormalization


def test_vectorspace_orthonormalization():
    a1 = np.array([[2, 0, 4, 12], [0, 2, 16, 4], [4, 16, 6, 2], [2, -12, 4, 6]])

    expect1 = np.array(
        [
            [0.40824829, -0.1814885, 0.04982278, 0.89325973],
            [0.0, 0.1088931, 0.99349591, -0.03328918],
            [0.81649658, 0.50816781, -0.06462163, -0.26631346],
            [0.40824829, -0.83484711, 0.07942048, -0.36063281],
        ]
    )

    np.testing.assert_array_almost_equal(
        vectorspace_orthonormalization(a1), expect1, decimal=7
    )

    np.testing.assert_array_almost_equal(
        vectorspace_orthonormalization(a1 / 2), expect1, decimal=7
    )


def test_vectorspace_dimensionality():
    a = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
    assert vectorspace_dimensionality(a) == 1

    a[:, 1] *= 2
    a[:, 2] *= 4

    assert vectorspace_dimensionality(a) == 1

    a[:, 1] += np.array([3, 34, 99])

    assert vectorspace_dimensionality(a) == 2

    b = np.array([[1, 2, 3], [15, 1, 3]])
    assert vectorspace_dimensionality(b) == 2



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Base Cluster (Clutering Parent Class)
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from time import time

import numpy as np


class _Cluster(object):
    def __init__(self):
        pass

    def fit(self, X, init_params=True):
        """Learn model from training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
        init_params : bool (default: True)
            Re-initializes model parameters prior to fitting.
            Set False to continue training with weights from
            a previous model fitting.

        Returns
        -------
        self : object

        """
        self._is_fitted = False
        self._check_arrays(X=X)
        if hasattr(self, "self.random_seed") and self.random_seed:
            self._rgen = np.random.RandomState(self.random_seed)
        self._init_time = time()
        self._fit(X=X, init_params=init_params)
        self._is_fitted = True
        return self

    def predict(self, X):
        """Predict targets from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        target_values : array-like, shape = [n_samples]
          Predicted target values.

        """
        self._check_arrays(X=X)
        if not self._is_fitted:
            raise AttributeError("Model is not fitted, yet.")
        return self._predict(X)



from collections import defaultdict

# The :mod:`sklearn.pipeline` module implements utilities to build a composite
# estimator, as a chain of transforms and estimators.
# Author: Edouard Duchesnay
#         Gael Varoquaux
#         Virgile Fritsch
#         Alexandre Gramfort
#         Lars Buitinck
# Licence: BSD


def _name_estimators(estimators):
    """Generate names for estimators."""

    names = [type(estimator).__name__.lower() for estimator in estimators]
    namecount = defaultdict(int)
    for _, name in zip(estimators, names):
        namecount[name] += 1

    for k, v in list(namecount.items()):
        if v == 1:
            del namecount[k]

    for i in reversed(range(len(estimators))):
        name = names[i]
        if name in namecount:
            names[i] += "-%d" % namecount[name]
            namecount[name] -= 1

    return list(zip(names, estimators))



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for plotting learning curves of classifiers.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np


def plot_learning_curves(
    X_train,
    y_train,
    X_test,
    y_test,
    clf,
    train_marker="o",
    test_marker="^",
    scoring="misclassification error",
    suppress_plot=False,
    print_model=True,
    title_fontsize=12,
    style="default",
    legend_loc="best",
):
    """Plots learning curves of a classifier.

    Parameters
    ----------
    X_train : array-like, shape = [n_samples, n_features]
        Feature matrix of the training dataset.
    y_train : array-like, shape = [n_samples]
        True class labels of the training dataset.
    X_test : array-like, shape = [n_samples, n_features]
        Feature matrix of the test dataset.
    y_test : array-like, shape = [n_samples]
        True class labels of the test dataset.
    clf : Classifier object. Must have a .predict .fit method.
    train_marker : str (default: 'o')
        Marker for the training set line plot.
    test_marker : str (default: '^')
        Marker for the test set line plot.
    scoring : str (default: 'misclassification error')
        If not 'misclassification error', accepts the following metrics
        (from scikit-learn):
        {'accuracy', 'average_precision', 'f1_micro', 'f1_macro',
        'f1_weighted', 'f1_samples', 'log_loss',
        'precision', 'recall', 'roc_auc',
        'adjusted_rand_score', 'mean_absolute_error', 'mean_squared_error',
        'median_absolute_error', 'r2'}
    suppress_plot=False : bool (default: False)
        Suppress matplotlib plots if True. Recommended
        for testing purposes.
    print_model : bool (default: True)
        Print model parameters in plot title if True.
    title_fontsize : int (default: 12)
        Determines the size of the plot title font.
    style : str (default: 'default')
        Matplotlib style. For more styles, please see
        https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
    legend_loc : str (default: 'best')
        Where to place the plot legend:
        {'best', 'upper left', 'upper right', 'lower left', 'lower right'}

    Returns
    ---------
    errors : (training_error, test_error): tuple of lists

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/plot_learning_curves/

    """
    if scoring != "misclassification error":
        from sklearn import metrics

        scoring_func = {
            "accuracy": metrics.accuracy_score,
            "average_precision": metrics.average_precision_score,
            "f1": metrics.f1_score,
            "f1_micro": metrics.f1_score,
            "f1_macro": metrics.f1_score,
            "f1_weighted": metrics.f1_score,
            "f1_samples": metrics.f1_score,
            "log_loss": metrics.log_loss,
            "precision": metrics.precision_score,
            "recall": metrics.recall_score,
            "roc_auc": metrics.roc_auc_score,
            "adjusted_rand_score": metrics.adjusted_rand_score,
            "mean_absolute_error": metrics.mean_absolute_error,
            "mean_squared_error": metrics.mean_squared_error,
            "median_absolute_error": metrics.median_absolute_error,
            "r2": metrics.r2_score,
        }

        if scoring not in scoring_func.keys():
            raise AttributeError("scoring must be in", scoring_func.keys())

    else:

        def misclf_err(y_predict, y):
            return (y_predict != y).sum() / float(len(y))

        scoring_func = {"misclassification error": misclf_err}

    training_errors = []
    test_errors = []

    rng = [int(i) for i in np.linspace(0, X_train.shape[0], 11)][1:]
    for r in rng:
        model = clf.fit(X_train[:r], y_train[:r])

        y_train_predict = clf.predict(X_train[:r])
        y_test_predict = clf.predict(X_test)

        train_misclf = scoring_func[scoring](y_train[:r], y_train_predict)
        training_errors.append(train_misclf)

        test_misclf = scoring_func[scoring](y_test, y_test_predict)
        test_errors.append(test_misclf)

    if not suppress_plot:
        with plt.style.context(style):
            plt.plot(
                np.arange(10, 101, 10),
                training_errors,
                label="training set",
                marker=train_marker,
            )
            plt.plot(
                np.arange(10, 101, 10),
                test_errors,
                label="test set",
                marker=test_marker,
            )
            plt.xlabel("Training set size in percent")

    if not suppress_plot:
        with plt.style.context(style):
            plt.ylabel("Performance ({})".format(scoring))
            if print_model:
                plt.title(
                    "Learning Curves\n\n{}\n".format(model), fontsize=title_fontsize
                )
            plt.legend(loc=legend_loc, numpoints=1)
            plt.xlim([0, 110])
            max_y = max(max(test_errors), max(training_errors))
            min_y = min(min(test_errors), min(training_errors))
            plt.ylim([min_y - min_y * 0.15, max_y + max_y * 0.15])
    errors = (training_errors, test_errors)
    return errors



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


from mlxtend.math import factorial, num_combinations, num_permutations


def test_factorial():
    assert factorial(1) == 1
    assert factorial(3) == 6


def test_num_combinations():
    assert num_combinations(n=20, k=8, with_replacement=False) == 125970
    assert num_combinations(n=20, k=8, with_replacement=True) == 2220075
    print(num_combinations(n=300, k=10))
    assert num_combinations(n=300, k=10, with_replacement=False) == 1398320233241701770


def test_num_permutations():
    assert num_permutations(n=20, k=8, with_replacement=False) == 5079110400
    assert num_permutations(n=20, k=8, with_replacement=True) == 25600000000



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from mlxtend.utils import Counter


def test_counter():
    cnt = Counter()
    for i in range(20):
        cnt.update()
    assert cnt.curr_iter == 20



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from numpy.testing import assert_almost_equal

from mlxtend.evaluate import mcnemar
from mlxtend.utils import assert_raises


def test_input_dimensions():
    t = np.ones((3, 3))
    assert_raises(ValueError, "Input array must be a 2x2 array.", mcnemar, t)


def test_defaults():
    tb = np.array([[101, 121], [59, 33]])

    chi2, p = (20.672222222222221, 5.4500948254271171e-06)
    chi2p, pp = mcnemar(tb)

    assert_almost_equal(chi2, chi2p, decimal=7)
    assert_almost_equal(p, pp, decimal=7)


def test_corrected_false():
    tb = np.array([[101, 121], [59, 33]])
    chi2, p = (21.355555555555554, 3.8151358651125936e-06)
    chi2p, pp = mcnemar(tb, corrected=False)

    assert_almost_equal(chi2, chi2p, decimal=7)
    assert_almost_equal(p, pp, decimal=7)


def test_exact():
    tb = np.array([[101, 121], [59, 33]])

    p = 4.4344492637555101e-06
    chi2p, pp = mcnemar(tb, exact=True)

    assert chi2p is None
    assert_almost_equal(p, pp, decimal=7)
    assert p < 4.45e-06


def test_exact_corrected():
    tb = np.array([[101, 121], [59, 33]])

    p = 4.4344492637555101e-06
    chi2p, pp = mcnemar(tb, exact=True, corrected=False)

    assert chi2p is None
    assert_almost_equal(p, pp, decimal=7)
    assert p < 4.45e-06



# Sebastian Raschka 2014-2016
# mlxtend Machine Learning Library Extensions
#
# A function for generating a confusion matrix.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


def lift_score(y_target, y_predicted, binary=True, positive_label=1):
    """Lift measures the degree to which the predictions of a
    classification model are better than randomly-generated predictions.

    The in terms of True Positives (TP), True Negatives (TN),
    False Positives (FP), and False Negatives (FN), the lift score is
    computed as:
    [ TP / (TP+FP) ] / [ (TP+FN) / (TP+TN+FP+FN) ]


    Parameters
    -----------
    y_target : array-like, shape=[n_samples]
        True class labels.
    y_predicted : array-like, shape=[n_samples]
        Predicted class labels.
    binary : bool (default: True)
        Maps a multi-class problem onto a
        binary, where
        the positive class is 1 and
        all other classes are 0.
    positive_label : int (default: 0)
        Class label of the positive class.

    Returns
    ----------
    score : float
        Lift score in the range [0, infinity]

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/lift_score/
    """
    if not isinstance(y_target, np.ndarray):
        targ_tmp = np.asarray(y_target)
    else:
        targ_tmp = y_target
    if not isinstance(y_predicted, np.ndarray):
        pred_tmp = np.asarray(y_predicted)
    else:
        pred_tmp = y_predicted

    pred_tmp = pred_tmp.T
    targ_tmp = targ_tmp.T

    if len(pred_tmp) != len(targ_tmp):
        raise AttributeError(
            "`y_target` and `y_predicted`" "don't have the same number of elements."
        )
    if binary:
        targ_tmp = np.where(targ_tmp != positive_label, 0, 1)
        pred_tmp = np.where(pred_tmp != positive_label, 0, 1)

    binary_check_targ_tmp = np.extract(targ_tmp > 1, targ_tmp)
    binary_check_pred_tmp = np.extract(pred_tmp > 1, pred_tmp)

    if len(binary_check_targ_tmp) or len(binary_check_pred_tmp):
        raise AttributeError(
            "`y_target` and `y_predicted`" " have different elements from 0 and 1."
        )

    return support(targ_tmp, pred_tmp) / (support(targ_tmp) * support(pred_tmp))


def support(y_target, y_predicted=None):
    """Support is the fraction of the true value
        in predictions and target values.

    Parameters
    -----------
    y_target : array-like, shape=[n_samples]
        True class labels.
    y_predicted : array-like, shape=[n_samples]
        Predicted class labels.

    Returns
    ----------
    score : float
        Support score in the range [0, 1]

    """

    if y_predicted is None:
        if y_target.ndim == 1:
            return (y_target == 1).sum() / float(y_target.shape[0])
        return (y_target == 1).all(axis=1).sum() / float(y_target.shape[0])
    else:
        all_prod = np.column_stack([y_target, y_predicted])
        return (all_prod == 1).all(axis=1).sum() / float(all_prod.shape[0])



print("import: 'mlxtend'")
import mlxtend

print("import: 'mlxtend._base'")
import mlxtend._base

print("import: 'mlxtend.classifier'")
import mlxtend.classifier

print("import: 'mlxtend.cluster'")
import mlxtend.cluster

print("import: 'mlxtend.data'")
import mlxtend.data

print("import: 'mlxtend.evaluate'")
import mlxtend.evaluate

print("import: 'mlxtend.externals'")
import mlxtend.externals

print("import: 'mlxtend.feature_extraction'")
import mlxtend.feature_extraction

print("import: 'mlxtend.feature_selection'")
import mlxtend.feature_selection

print("import: 'mlxtend.file_io'")
import mlxtend.file_io

print("import: 'mlxtend.frequent_patterns'")
import mlxtend.frequent_patterns

print("import: 'mlxtend.math'")
import mlxtend.math

print("import: 'mlxtend.plotting'")
import mlxtend.plotting

print("import: 'mlxtend.preprocessing'")
import mlxtend.preprocessing

print("import: 'mlxtend.regressor'")
import mlxtend.regressor

print("import: 'mlxtend.text'")
import mlxtend.text

print("import: 'mlxtend.utils'")
import mlxtend.utils




# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from .apriori import apriori
from .association_rules import association_rules
from .fpgrowth import fpgrowth
from .fpmax import fpmax
from .hmine import hmine

__all__ = ["apriori", "association_rules", "fpgrowth", "fpmax", "hmine"]



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import itertools

import numpy as np
import scipy.stats


def cochrans_q(y_target, *y_model_predictions):
    """
    Cochran's Q test to compare 2 or more models.

    Parameters
    -----------
    y_target : array-like, shape=[n_samples]
        True class labels as 1D NumPy array.

    *y_model_predictions : array-likes, shape=[n_samples]
        Variable number of 2 or more arrays that
        contain the predicted class labels
        from models as 1D NumPy array.

    Returns
    -----------

    q, p : float or None, float
        Returns the Q (chi-squared) value and the p-value

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/cochrans_q/

    """

    num_models = len(y_model_predictions)

    # Checks
    model_lens = set()
    y_model_predictions = list(y_model_predictions)
    for ary in [y_target] + y_model_predictions:
        if len(ary.shape) != 1:
            raise ValueError("One or more input arrays are not 1-dimensional.")
        model_lens.add(ary.shape[0])

    if len(model_lens) > 1:
        raise ValueError(
            "Each prediction array must have the " "same number of samples."
        )

    if num_models < 2:
        raise ValueError("Provide at least 2 model prediction arrays.")

    # Q test statistic
    degrees_of_freedom = num_models - 1

    # numerator
    correctly_classified_all_models = 0
    correctly_classified_collection = []
    for pred in y_model_predictions:
        correctly_classified = (y_target == pred).sum()
        correctly_classified_all_models += correctly_classified
        correctly_classified_collection.append(correctly_classified)

    numerator = (
        num_models * sum([c**2 for c in correctly_classified_collection])
        - correctly_classified_all_models**2
    )

    # denominator
    binary_combin = list(itertools.product([0, 1], repeat=num_models))
    ary = np.hstack(
        [(y_target == mod).reshape(-1, 1) for mod in y_model_predictions]
    ).astype(int)
    correctly_classified_objects = 0
    binary_combin_totals = np.zeros(len(binary_combin))
    for i, c in enumerate(binary_combin):
        binary_combin_totals[i] = ((ary == c).sum(axis=1) == num_models).sum()

        correctly_classified_objects += sum(c) ** 2 * binary_combin_totals[i]

    denominator = (
        num_models * correctly_classified_all_models - correctly_classified_objects
    )

    q = degrees_of_freedom * numerator / denominator
    p_value = scipy.stats.chi2.sf(q, degrees_of_freedom)

    return q, p_value



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for generating per class accuracy
#
# License: BSD 3 clause
import numpy as np


def _compute_metric(target_temp, predicted_temp, normalize=True):
    if normalize:
        return (target_temp == predicted_temp).mean()
    else:
        return (target_temp == predicted_temp).sum()


def accuracy_score(
    y_target, y_predicted, method="standard", pos_label=1, normalize=True
):
    """General accuracy function for supervised learning.
    Parameters
    ------------
    y_target : array-like, shape=[n_values]
        True class labels or target values.

    y_predicted : array-like, shape=[n_values]
        Predicted class labels or target values.

    method : str, 'standard' by default.
        The chosen method for accuracy computation.
        If set to 'standard', computes overall accuracy.
        If set to 'binary', computes accuracy for class pos_label.
        If set to 'average', computes average per-class (balanced) accuracy.
        If set to 'balanced', computes the scikit-learn-style balanced accuracy.

    pos_label : str or int, 1 by default.
        The class whose accuracy score is to be reported.
        Used only when `method` is set to 'binary'

    normalize : bool, True by default.
        If True, returns fraction of correctly classified samples.
        If False, returns number of correctly classified samples.

    Returns
    ------------
    score: float

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/accuracy_score/

    """

    target_temp = np.asarray(y_target)
    predicted_temp = np.asarray(y_predicted)
    unique_labels = np.unique(target_temp)

    if len(y_target) != len(y_predicted):
        raise AttributeError(
            "`y_target` and `y_predicted`" " don't have the same number of elements."
        )
    if method == "standard":
        return _compute_metric(target_temp, predicted_temp, normalize)

    elif method == "binary":
        if pos_label not in unique_labels:
            raise AttributeError("Chosen value of pos_label doesn't exist")

        target_temp = np.where(target_temp == pos_label, 1, 0)
        predicted_temp = np.where(predicted_temp == pos_label, 1, 0)
        return _compute_metric(target_temp, predicted_temp, normalize)

    elif method == "average":
        return sum(
            [
                _compute_metric(
                    np.where(target_temp != lab, 1, 0),
                    np.where(predicted_temp != lab, 1, 0),
                )
                for lab in unique_labels
            ]
        ) / float(unique_labels.shape[0])

    elif method == "balanced":
        all_class_acc = []
        for c in np.unique(y_target):
            positive_labels = y_target == c
            class_acc = np.mean((y_predicted == c)[positive_labels])
            all_class_acc.append(class_acc)

        return np.mean(all_class_acc)

    else:
        raise ValueError(
            '`method` must be "standard", "average",'
            ' "balanced", or "binary". Got "%s".' % method
        )



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Authors: Sebastian Raschka <sebastianraschka.com>
#          Reiichiro Nakano <github.com/reiinakano>
#
# License: BSD 3 clause

import random

import numpy as np
import pandas as pd
import pytest
from packaging.version import Version
from scipy import sparse
from sklearn import __version__ as sklearn_version
from sklearn import datasets, exceptions
from sklearn.base import clone
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression, PassiveAggressiveClassifier
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import (
    GridSearchCV,
    KFold,
    cross_val_score,
    train_test_split,
)
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC

from mlxtend.classifier import StackingCVClassifier
from mlxtend.data import iris_data
from mlxtend.externals.estimator_checks import NotFittedError
from mlxtend.utils import assert_raises

X_iris, y_iris = iris_data()
X_iris = X_iris[:, 1:3]

breast_cancer = datasets.load_breast_cancer()
X_breast, y_breast = breast_cancer.data[:, 1:3], breast_cancer.target


def test_StackingCVClassifier():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, shuffle=False
    )

    scores = cross_val_score(sclf, X_iris, y_iris, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)

    if Version(sklearn_version) < Version("0.20"):
        assert scores_mean == 0.93, scores_mean
    else:
        assert scores_mean == 0.92, scores_mean


def test_use_clones():
    np.random.seed(123)
    X, y = iris_data()

    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    StackingCVClassifier(
        classifiers=[clf1, clf2], use_clones=True, meta_classifier=meta, shuffle=False
    ).fit(X, y)

    assert_raises(
        exceptions.NotFittedError,
        "This RandomForestClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this estimator.",
        clf1.predict,
        X,
    )

    StackingCVClassifier(
        classifiers=[clf1, clf2],
        use_probas=True,
        use_clones=False,
        meta_classifier=meta,
        shuffle=False,
    ).fit(X, y)

    clf1.predict(X)


def test_sample_weight():
    # with no weight given
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, shuffle=False
    )
    prob1 = sclf.fit(X_iris, y_iris).predict_proba(X_iris)

    # with weight = 1
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, shuffle=False
    )
    w = np.ones(len(y_iris))
    prob2 = sclf.fit(X_iris, y_iris, sample_weight=w).predict_proba(X_iris)

    # with random weight
    random.seed(87)
    w = np.array([random.random() for _ in range(len(y_iris))])
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, shuffle=False
    )
    prob3 = sclf.fit(X_iris, y_iris, sample_weight=w).predict_proba(X_iris)

    diff12 = np.max(np.abs(prob1 - prob2))
    diff23 = np.max(np.abs(prob2 - prob3))
    assert diff12 < 1e-3, "max diff is %.4f" % diff12
    assert diff23 > 1e-3, "max diff is %.4f" % diff23


def test_no_weight_support():
    w = np.array([random.random() for _ in range(len(y_iris))])
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    clf3 = KNeighborsClassifier()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2, clf3], meta_classifier=meta, shuffle=False
    )
    with pytest.raises(TypeError):
        sclf.fit(X_iris, y_iris, sample_weight=w)


def test_no_weight_support_meta():
    w = np.array([random.random() for _ in range(len(y_iris))])
    meta = KNeighborsClassifier()
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, shuffle=False
    )

    with pytest.raises(TypeError):
        sclf.fit(X_iris, y_iris, sample_weight=w)


def test_no_weight_support_with_no_weight():
    logit = LogisticRegression(multi_class="ovr", solver="liblinear")
    rf = RandomForestClassifier(n_estimators=10)
    gnb = GaussianNB()
    knn = KNeighborsClassifier()
    sclf = StackingCVClassifier(
        classifiers=[logit, rf, gnb], meta_classifier=knn, shuffle=False
    )
    sclf.fit(X_iris, y_iris)

    sclf = StackingCVClassifier(
        classifiers=[logit, knn, gnb], meta_classifier=rf, shuffle=False
    )
    sclf.fit(X_iris, y_iris)


def test_StackingClassifier_proba():
    np.random.seed(12)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, shuffle=False
    )

    scores = cross_val_score(sclf, X_iris, y_iris, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)

    if Version(sklearn_version) < Version("0.20"):
        assert scores_mean == 0.92, scores_mean
    else:
        assert scores_mean == 0.93, scores_mean


def test_gridsearch():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, use_probas=True, shuffle=False
    )

    params = {
        "meta_classifier__C": [1.0, 100.0],
        "randomforestclassifier__n_estimators": [20, 200],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5, iid=False)
    else:
        grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5)
    X, y = iris_data()
    grid.fit(X, y)

    mean_scores = [round(s, 2) for s in grid.cv_results_["mean_test_score"]]

    assert mean_scores == [0.96, 0.95, 0.96, 0.95]


def test_gridsearch_enumerate_names():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf1, clf2], meta_classifier=meta, shuffle=False
    )

    params = {
        "meta_classifier__C": [1.0, 100.0],
        "randomforestclassifier-1__n_estimators": [5, 10],
        "randomforestclassifier-2__n_estimators": [5, 20],
        "use_probas": [True, False],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5, iid=False)
    else:
        grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5)
    X, y = iris_data()
    grid = grid.fit(X, y)


def test_use_probas():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta, shuffle=False
    )

    scores = cross_val_score(sclf, X_iris, y_iris, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.94, scores_mean


def test_use_features_in_secondary():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2],
        use_features_in_secondary=True,
        meta_classifier=meta,
        shuffle=False,
    )

    scores = cross_val_score(sclf, X_iris, y_iris, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.93, scores_mean


def test_do_not_stratify():
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, random_state=42, stratify=False
    )

    scores = cross_val_score(sclf, X_iris, y_iris, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.93, scores.mean()


def test_cross_validation_technique():
    # This is like the `test_do_not_stratify` but instead
    # autogenerating the cross validation strategy it provides
    # a pre-created object
    cv = KFold(n_splits=2, shuffle=True)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10, random_state=42)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, cv=cv, random_state=42
    )

    scores = cross_val_score(sclf, X_iris, y_iris, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.92, scores.mean()


def test_not_fitted():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta, shuffle=False
    )

    X, y = iris_data()
    assert_raises(
        NotFittedError,
        "This StackingCVClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this method.",
        sclf.predict,
        X,
    )

    assert_raises(
        NotFittedError,
        "This StackingCVClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this method.",
        sclf.predict_proba,
        X,
    )

    assert_raises(
        NotFittedError,
        "This StackingCVClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this method.",
        sclf.predict_meta_features,
        X,
    )


def test_verbose():
    np.random.seed(123)
    meta = LogisticRegression(multi_class="ovr", solver="liblinear")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2],
        use_probas=True,
        meta_classifier=meta,
        shuffle=False,
        verbose=3,
    )
    sclf.fit(X_iris, y_iris)


def test_get_params():
    clf1 = KNeighborsClassifier(n_neighbors=1)
    clf2 = RandomForestClassifier(random_state=1)
    clf3 = GaussianNB()
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    sclf = StackingCVClassifier(classifiers=[clf1, clf2, clf3], meta_classifier=lr)

    got = sorted(list({s.split("__")[0] for s in sclf.get_params().keys()}))

    expect = [
        "classifiers",
        "cv",
        "drop_proba_col",
        "gaussiannb",
        "kneighborsclassifier",
        "meta_classifier",
        "n_jobs",
        "pre_dispatch",
        "random_state",
        "randomforestclassifier",
        "shuffle",
        "store_train_meta_features",
        "stratify",
        "use_clones",
        "use_features_in_secondary",
        "use_probas",
        "verbose",
    ]
    assert got == expect, got


def test_classifier_gridsearch():
    clf1 = KNeighborsClassifier(n_neighbors=1)
    clf2 = RandomForestClassifier(n_estimators=10, random_state=42)
    clf3 = GaussianNB()
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    sclf = StackingCVClassifier(classifiers=[clf1], meta_classifier=lr, random_state=42)

    params = {"classifiers": [[clf1], [clf1, clf2, clf3]]}

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=sclf, param_grid=params, cv=5, iid=False, refit=True
        )
    else:
        grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5, refit=True)
    grid.fit(X_iris, y_iris)

    assert len(grid.best_params_["classifiers"]) == 3, len(
        grid.best_params_["classifiers"]
    )


def test_train_meta_features_():
    knn = KNeighborsClassifier()
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    gnb = GaussianNB()
    stclf = StackingCVClassifier(
        classifiers=[knn, gnb], meta_classifier=lr, store_train_meta_features=True
    )
    X_train, _, y_train, _ = train_test_split(X_iris, y_iris, test_size=0.3)
    stclf.fit(X_train, y_train)
    train_meta_features = stclf.train_meta_features_
    assert train_meta_features.shape == (X_train.shape[0], 2)


def test_predict_meta_features():
    knn = KNeighborsClassifier()
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    gnb = GaussianNB()
    X_train, X_test, y_train, y_test = train_test_split(X_iris, y_iris, test_size=0.3)
    #  test default (class labels)
    stclf = StackingCVClassifier(
        classifiers=[knn, gnb], meta_classifier=lr, store_train_meta_features=True
    )
    stclf.fit(X_train, y_train)
    test_meta_features = stclf.predict(X_test)
    assert test_meta_features.shape == (X_test.shape[0],)


def test_meta_feat_reordering():
    knn = KNeighborsClassifier()
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    gnb = GaussianNB()
    stclf = StackingCVClassifier(
        classifiers=[knn, gnb],
        meta_classifier=lr,
        shuffle=True,
        random_state=42,
        store_train_meta_features=True,
    )
    X_train, _, y_train, _ = train_test_split(
        X_breast, y_breast, random_state=0, test_size=0.3
    )
    stclf.fit(X_train, y_train)

    if Version(sklearn_version) < Version("0.21"):
        expected_value = 0.86
    elif Version(sklearn_version) < Version("0.22"):
        expected_value = 0.87
    else:
        expected_value = 0.85

    assert (
        round(roc_auc_score(y_train, stclf.train_meta_features_[:, 1]), 2)
        == expected_value
    ), round(roc_auc_score(y_train, stclf.train_meta_features_[:, 1]), 2)


def test_clone():
    knn = KNeighborsClassifier()
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    gnb = GaussianNB()
    stclf = StackingCVClassifier(
        classifiers=[knn, gnb], meta_classifier=lr, store_train_meta_features=True
    )
    clone(stclf)


def test_sparse_inputs():
    np.random.seed(123)
    rf = RandomForestClassifier(n_estimators=10)
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    stclf = StackingCVClassifier(
        classifiers=[rf, rf], meta_classifier=lr, random_state=42
    )
    X_train, X_test, y_train, y_test = train_test_split(
        X_breast, y_breast, test_size=0.3
    )

    # dense
    stclf.fit(X_train, y_train)
    assert round(stclf.score(X_train, y_train), 2) == 0.99

    # sparse
    stclf.fit(sparse.csr_matrix(X_train), y_train)
    assert round(stclf.score(X_train, y_train), 2) == 0.99


def test_sparse_inputs_with_features_in_secondary():
    rf = RandomForestClassifier(n_estimators=10, random_state=42)
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    stclf = StackingCVClassifier(
        classifiers=[rf, rf],
        meta_classifier=lr,
        random_state=42,
        use_features_in_secondary=True,
    )
    X_train, _, y_train, _ = train_test_split(X_breast, y_breast, test_size=0.3)

    # dense
    stclf.fit(X_train, y_train)

    expected_value = 1.0

    assert round(stclf.score(X_train, y_train), 2) == expected_value, round(
        stclf.score(X_train, y_train), 2
    )

    # sparse
    stclf.fit(sparse.csr_matrix(X_train), y_train)

    if Version(sklearn_version) < Version("0.21"):
        expected_value = 1.0
    if Version(sklearn_version) < Version("0.22"):
        expected_value = 0.99
    else:
        expected_value = 1.00

    assert round(stclf.score(X_train, y_train), 2) == expected_value, round(
        stclf.score(X_train, y_train), 2
    )


def test_StackingClassifier_drop_proba_col():
    np.random.seed(123)
    lr1 = LogisticRegression(solver="liblinear", multi_class="ovr")
    sclf1 = StackingCVClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col=None,
        meta_classifier=lr1,
    )

    sclf1.fit(X_iris, y_iris)
    r1 = sclf1.predict_meta_features(X_iris[:2])
    assert r1.shape == (2, 6)

    sclf2 = StackingCVClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col="last",
        meta_classifier=lr1,
    )

    sclf2.fit(X_iris, y_iris)
    r2 = sclf2.predict_meta_features(X_iris[:2])
    assert r2.shape == (2, 4), r2.shape

    sclf4 = StackingCVClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col="first",
        meta_classifier=lr1,
    )

    sclf4.fit(X_iris, y_iris)
    r4 = sclf4.predict_meta_features(X_iris[:2])
    assert r4.shape == (2, 4), r4.shape

    sclf3 = StackingCVClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col="last",
        meta_classifier=lr1,
    )

    sclf3.fit(X_iris[0:100], y_iris[0:100])  # only 2 classes
    r3 = sclf3.predict_meta_features(X_iris[:2])
    assert r3.shape == (2, 2), r3.shape


def test_works_with_df_if_fold_indexes_missing():
    """This is a regression test to make sure fitting will still work even if
    training data has ids that cannot be indexed using the indexes from the cv
    (e.g. skf)

    Some possibilities:
    + Output of the folds are not neatly consecutive (i.e. [341, 345, 543, ...]
      instead of [0, 1, ... n])
    + Indexes just start from some number greater than the size of the input
      (see test case)

    Training data sometimes has ids that carry other information, and selection
    of rows based on cv should not break.

    This is fixed in the code using `safe_indexing`
    """

    np.random.seed(123)
    rf = RandomForestClassifier(n_estimators=10, random_state=42)
    lr = LogisticRegression(multi_class="ovr", solver="liblinear")
    stclf = StackingCVClassifier(
        classifiers=[rf, rf],
        meta_classifier=lr,
        random_state=42,
        use_features_in_secondary=True,
    )

    X_modded = pd.DataFrame(X_breast, index=np.arange(X_breast.shape[0]) + 1000)
    y_modded = pd.Series(y_breast, index=np.arange(y_breast.shape[0]) + 1000)

    X_train, X_test, y_train, y_test = train_test_split(
        X_modded, y_modded, test_size=0.3
    )

    # dense
    stclf.fit(X_train, y_train)

    if Version(sklearn_version) < Version("0.22"):
        assert round(stclf.score(X_train, y_train), 2) == 0.99, round(
            stclf.score(X_train, y_train), 2
        )
    else:
        assert round(stclf.score(X_train, y_train), 2) == 0.98, round(
            stclf.score(X_train, y_train), 2
        )


def test_decision_function():
    np.random.seed(123)

    # PassiveAggressiveClassifier has no predict_proba
    meta = PassiveAggressiveClassifier(random_state=42)
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, X_breast, y_breast, cv=5, scoring="roc_auc")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.96, scores_mean

    # another test
    meta = SVC(decision_function_shape="ovo")

    sclf = StackingCVClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, X_breast, y_breast, cv=5, scoring="roc_auc")
    scores_mean = round(scores.mean(), 2)

    if Version(sklearn_version) < Version("0.21"):
        assert scores_mean == 0.94, scores_mean
    elif Version(sklearn_version) < Version("0.22"):
        assert scores_mean == 0.96, scores_mean
    else:
        assert scores_mean == 0.90, scores_mean


def test_drop_col_unsupported():
    np.random.seed(123)
    meta = LogisticRegression()
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    clf3 = KNeighborsClassifier()

    with pytest.raises(ValueError):
        StackingCVClassifier(
            classifiers=[clf1, clf2, clf3],
            meta_classifier=meta,
            drop_proba_col="invalid value",
        )



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Base Iterative Model (Iterative Model Parent Class)
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from sys import stderr
from time import time

import numpy as np


class _IterativeModel(object):
    def __init__(self):
        pass

    def _shuffle_arrays(self, arrays):
        """Shuffle arrays in unison."""
        r = np.random.permutation(len(arrays[0]))
        return [ary[r] for ary in arrays]

    def _print_progress(self, iteration, n_iter, cost=None, time_interval=10):
        if self.print_progress > 0:
            s = "\rIteration: %d/%d" % (iteration, n_iter)
            if cost:
                s += " | Cost %.2f" % cost
            if self.print_progress > 1:
                if not hasattr(self, "ela_str_"):
                    self.ela_str_ = "00:00:00"
                if not iteration % time_interval:
                    ela_sec = time() - self._init_time
                    self.ela_str_ = self._to_hhmmss(ela_sec)
                s += " | Elapsed: %s" % self.ela_str_
                if self.print_progress > 2:
                    if not hasattr(self, "eta_str_"):
                        self.eta_str_ = "00:00:00"
                    if not iteration % time_interval:
                        eta_sec = (ela_sec / float(iteration)) * n_iter - ela_sec
                        if eta_sec < 0.0:
                            eta_sec = 0.0
                        self.eta_str_ = self._to_hhmmss(eta_sec)
                    s += " | ETA: %s" % self.eta_str_
            stderr.write(s)
            stderr.flush()

    def _to_hhmmss(self, sec):
        m, s = divmod(sec, 60)
        h, m = divmod(m, 60)
        return "%d:%02d:%02d" % (h, m, s)

    def _yield_minibatches_idx(self, rgen, n_batches, data_ary, shuffle=True):
        indices = np.arange(data_ary.shape[0])

        if shuffle:
            indices = rgen.permutation(indices)
        if n_batches > 1:
            remainder = data_ary.shape[0] % n_batches

            if remainder:
                minis = np.array_split(indices[:-remainder], n_batches)
                minis[-1] = np.concatenate((minis[-1], indices[-remainder:]), axis=0)
            else:
                minis = np.array_split(indices, n_batches)

        else:
            minis = (indices,)

        for idx_batch in minis:
            yield idx_batch

    def _init_params(
        self,
        weights_shape,
        bias_shape=(1,),
        random_seed=None,
        dtype="float64",
        scale=0.01,
        bias_const=0.0,
    ):
        """Initialize weight coefficients."""
        rgen = np.random.RandomState(random_seed)
        w = rgen.normal(loc=0.0, scale=scale, size=weights_shape)
        b = np.zeros(shape=bias_shape)
        if bias_const != 0.0:
            b += bias_const
        return b.astype(dtype), w.astype(dtype)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Feature Importance Estimation Through Permutation
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification, make_regression
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from sklearn.svm import SVC, SVR

from mlxtend.evaluate import feature_importance_permutation
from mlxtend.utils import assert_raises


def test_num_rounds_not_int():
    assert_raises(
        ValueError,
        "num_rounds must be an integer.",
        feature_importance_permutation,
        lambda x, y: (x, y),
        np.array([[1], [2], [3]]),
        np.array([1, 2, 3]),
        "accuracy",
        1.23,
    )


def test_num_rounds_negative_int():
    assert_raises(
        ValueError,
        "num_rounds must be greater than 1.",
        feature_importance_permutation,
        lambda x, y: (x, y),
        np.array([[1], [2], [3]]),
        np.array([1, 2, 3]),
        "accuracy",
        -1,
    )


def test_metric_wrong():
    assert_raises(
        ValueError,
        (
            'metric must be either "r2", "accuracy", or a '
            "function with signature "
            "func(y_true, y_pred)."
        ),
        feature_importance_permutation,
        lambda x, y: (x, y),
        np.array([[1], [2], [3]]),
        np.array([1, 2, 3]),
        "some-metric",
    )


def test_classification():
    X, y = make_classification(
        n_samples=1000,
        n_features=6,
        n_informative=3,
        n_redundant=0,
        n_repeated=0,
        n_classes=2,
        random_state=0,
        shuffle=False,
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=0, stratify=y
    )

    svm = SVC(C=1.0, kernel="rbf", random_state=0, gamma="auto")
    svm.fit(X_train, y_train)

    imp_vals, imp_all = feature_importance_permutation(
        predict_method=svm.predict,
        X=X_test,
        y=y_test,
        metric="accuracy",
        num_rounds=1,
        seed=1,
    )

    assert imp_vals.shape == (X_train.shape[1],)
    assert imp_all.shape == (X_train.shape[1], 1)
    assert imp_vals[0] > 0.2
    assert imp_vals[1] > 0.2
    assert imp_vals[2] > 0.2
    assert sum(imp_vals[3:]) <= 0.02


def test_regression():
    X, y = make_regression(
        n_samples=1000,
        n_features=5,
        n_informative=2,
        n_targets=1,
        random_state=123,
        shuffle=False,
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123
    )

    svm = SVR(kernel="rbf", gamma="auto")
    svm.fit(X_train, y_train)

    imp_vals, imp_all = feature_importance_permutation(
        predict_method=svm.predict,
        X=X_test,
        y=y_test,
        metric="r2",
        num_rounds=1,
        seed=123,
    )

    assert imp_vals.shape == (X_train.shape[1],)
    assert imp_all.shape == (X_train.shape[1], 1)
    assert imp_vals[0] > 0.2
    assert imp_vals[1] > 0.2
    assert sum(imp_vals[3:]) <= 0.01


def test_regression_custom_r2():
    X, y = make_regression(
        n_samples=1000,
        n_features=5,
        n_informative=2,
        n_targets=1,
        random_state=123,
        shuffle=False,
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123
    )

    svm = SVR(kernel="rbf", gamma="auto")
    svm.fit(X_train, y_train)

    imp_vals, imp_all = feature_importance_permutation(
        predict_method=svm.predict,
        X=X_test,
        y=y_test,
        metric=r2_score,
        num_rounds=1,
        seed=123,
    )

    assert imp_vals.shape == (X_train.shape[1],)
    assert imp_all.shape == (X_train.shape[1], 1)
    assert imp_vals[0] > 0.2
    assert imp_vals[1] > 0.2
    assert sum(imp_vals[3:]) <= 0.01


def test_regression_custom_mse():
    X, y = make_regression(
        n_samples=1000,
        n_features=5,
        n_informative=2,
        n_targets=1,
        random_state=123,
        shuffle=False,
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=123
    )

    svm = SVR(kernel="rbf", gamma="auto")
    svm.fit(X_train, y_train)

    imp_vals, imp_all = feature_importance_permutation(
        predict_method=svm.predict,
        X=X_test,
        y=y_test,
        metric=mean_squared_error,
        num_rounds=1,
        seed=123,
    )

    norm_imp_vals = imp_vals / np.abs(imp_vals).max()

    assert imp_vals.shape == (X_train.shape[1],)
    assert imp_all.shape == (X_train.shape[1], 1)
    assert norm_imp_vals[0] == -1.0


def test_n_rounds():
    X, y = make_classification(
        n_samples=1000,
        n_features=6,
        n_informative=3,
        n_redundant=0,
        n_repeated=0,
        n_classes=2,
        random_state=0,
        shuffle=False,
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=0, stratify=y
    )

    svm = SVC(C=1.0, kernel="rbf", random_state=0, gamma="auto")
    svm.fit(X_train, y_train)

    imp_vals, imp_all = feature_importance_permutation(
        predict_method=svm.predict,
        X=X_test,
        y=y_test,
        metric="accuracy",
        num_rounds=100,
        seed=1,
    )

    assert imp_vals.shape == (X_train.shape[1],)
    assert imp_all.shape == (X_train.shape[1], 100)
    assert imp_vals[0].mean() > 0.2
    assert imp_vals[1].mean() > 0.2


def test_feature_groups():
    df_data = pd.read_csv(
        "https://gist.githubusercontent.com/rasbt/"
        "b99bf69079bc0d601eeae8a49248d358/"
        "raw/a114be9801647ec5460089f3a9576713dabf5f1f/"
        "onehot-numeric-mixed-data.csv"
    )

    df_X = df_data[["measurement1", "measurement2", "measurement3", "categorical"]]
    df_y = df_data["label"]

    df_X_train, df_X_test, df_y_train, df_y_test = train_test_split(
        df_X, df_y, test_size=0.33, random_state=42, stratify=df_y
    )

    ohe = OneHotEncoder(drop="first")
    ohe.fit(df_X_train[["categorical"]])

    df_X_train_ohe = df_X_train.drop(columns=["categorical"])
    df_X_test_ohe = df_X_test.drop(columns=["categorical"])

    ohe_train = np.asarray(ohe.transform(df_X_train[["categorical"]]).todense())
    ohe_test = np.asarray(ohe.transform(df_X_test[["categorical"]]).todense())

    X_train_ohe = np.hstack((df_X_train_ohe.values, ohe_train))
    X_test_ohe = np.hstack((df_X_test_ohe.values, ohe_test))

    model = LogisticRegression().fit(X_train_ohe, df_y_train.values)

    feature_groups = [0, 1, 2, range(3, 21)]

    imp_vals, imp_all = feature_importance_permutation(
        predict_method=model.predict,
        X=X_test_ohe,
        y=df_y_test.values,
        metric="accuracy",
        num_rounds=50,
        feature_groups=feature_groups,
        seed=1,
    )

    assert len(imp_vals) == 4



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import os
import sys

import numpy as np

from mlxtend.utils import assert_raises, check_Xy, format_kwarg_dictionaries

y = np.array([1, 2, 3, 4])
X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])

d_default = {"key1": 1, "key2": 2}
d_user = {"key3": 3, "key4": 4}
protected_keys = ["key1", "key4"]


def test_check_Xy_ok():
    check_Xy(X, y)


def test_check_Xy_invalid_type_X():
    expect = "X must be a NumPy array. Found <class 'list'>"
    if sys.version_info < (3, 0):
        expect = expect.replace("class", "type")
    assert_raises(ValueError, expect, check_Xy, [1, 2, 3, 4], y)


def test_check_Xy_float16_X():
    check_Xy(X.astype(np.float16), y)


def test_check_Xy_float16_y():
    check_Xy(X, y.astype(np.int16))


def test_check_Xy_invalid_type_y():
    expect = "y must be a NumPy array. Found <class 'list'>"
    if sys.version_info < (3, 0):
        expect = expect.replace("class", "type")
    assert_raises(ValueError, expect, check_Xy, X, [1, 2, 3, 4])


def test_check_Xy_invalid_dtype_X():
    assert_raises(
        ValueError,
        "X must be an integer or float array. Found object.",
        check_Xy,
        X.astype("object"),
        y,
    )


def test_check_Xy_invalid_dtype_y():
    if sys.version_info > (3, 0):
        expect = (
            "y must be an integer array. Found <U1. "
            "Try passing the array as y.astype(np.int_)"
        )
    else:
        expect = (
            "y must be an integer array. Found |S1. "
            "Try passing the array as y.astype(np.int_)"
        )
    assert_raises(ValueError, expect, check_Xy, X, np.array(["a", "b", "c", "d"]))


def test_check_Xy_invalid_dim_y():
    if sys.version_info[:2] == (2, 7) and os.name == "nt":
        s = "y must be a 1D array. Found (4L, 2L)"
    else:
        s = "y must be a 1D array. Found (4, 2)"
    assert_raises(ValueError, s, check_Xy, X, X.astype(np.int_))


def test_check_Xy_invalid_dim_X():
    if sys.version_info[:2] == (2, 7) and os.name == "nt":
        s = "X must be a 2D array. Found (4L,)"
    else:
        s = "X must be a 2D array. Found (4,)"
    assert_raises(ValueError, s, check_Xy, y, y)


def test_check_Xy_unequal_length_X():
    assert_raises(
        ValueError,
        ("y and X must contain the same number of samples. " "Got y: 4, X: 3"),
        check_Xy,
        X[1:],
        y,
    )


def test_check_Xy_unequal_length_y():
    assert_raises(
        ValueError,
        ("y and X must contain the same number of samples. " "Got y: 3, X: 4"),
        check_Xy,
        X,
        y[1:],
    )


def test_format_kwarg_dictionaries_defaults_empty():
    empty = format_kwarg_dictionaries()
    assert isinstance(empty, dict)
    assert len(empty) == 0


def test_format_kwarg_dictionaries_protected_keys():
    formatted_kwargs = format_kwarg_dictionaries(
        default_kwargs=d_default, user_kwargs=d_user, protected_keys=protected_keys
    )

    for key in protected_keys:
        assert key not in formatted_kwargs


def test_format_kwarg_dictionaries_no_default_kwargs():
    formatted_kwargs = format_kwarg_dictionaries(user_kwargs=d_user)
    assert formatted_kwargs == d_user


def test_format_kwarg_dictionaries_no_user_kwargs():
    formatted_kwargs = format_kwarg_dictionaries(default_kwargs=d_default)
    assert formatted_kwargs == d_default


def test_format_kwarg_dictionaries_default_kwargs_invalid_type():
    invalid_kwargs = "not a dictionary"
    message = "d must be of type dict or None, but got " "{} instead".format(
        type(invalid_kwargs)
    )
    assert_raises(
        TypeError, message, format_kwarg_dictionaries, default_kwargs=invalid_kwargs
    )


def test_format_kwarg_dictionaries_user_kwargs_invalid_type():
    invalid_kwargs = "not a dictionary"
    message = "d must be of type dict or None, but got " "{} instead".format(
        type(invalid_kwargs)
    )
    assert_raises(
        TypeError, message, format_kwarg_dictionaries, user_kwargs=invalid_kwargs
    )



# mlxtend Machine Learning Library Extensions
# Author: Fatih Sen <fatih.sn2000@gmail.com>
#
# License: BSD 3 clause

import math

import numpy as np
import pandas as pd

from ..frequent_patterns import fpcommon as fpc


def hmine(
    df, min_support=0.5, use_colnames=False, max_len=None, verbose=0
) -> pd.DataFrame:
    """
    Get frequent itemsets from a one-hot DataFrame

    Parameters
    -----------
    df : pandas DataFrame
      pandas DataFrame the encoded format. Also supports
      DataFrames with sparse data; for more info, please
      see https://pandas.pydata.org/pandas-docs/stable/user_guide/sparse.html#sparse-data-structures.

      Please note that the old pandas SparseDataFrame format
      is no longer supported in mlxtend >= 0.17.2.

      The allowed values are either 0/1 or True/False.
      For example,

    ```
           Apple  Bananas   Beer  Chicken   Milk   Rice
        0   True    False   True     True  False   True
        1   True    False   True    False  False   True
        2   True    False   True    False  False  False
        3   True     True  False    False  False  False
        4  False    False   True     True   True   True
        5  False    False   True    False   True   True
        6  False    False   True    False   True  False
        7   True     True  False    False  False  False
    ```

    min_support : float (default: 0.5)
      A float between 0 and 1 for minimum support of the itemsets returned.
      The support is computed as the fraction
      transactions_where_item(s)_occur / total_transactions.

    use_colnames : bool (default: False)
      If true, uses the DataFrames' column names in the returned DataFrame
      instead of column indices.

    max_len : int (default: None)
      Maximum length of the itemsets generated. If `None` (default) all
      possible itemsets lengths are evaluated.

    verbose : int (default: 0)
      Shows the stages of conditional tree generation.

    Returns
    -----------
    pandas DataFrame with columns ['support', 'itemsets'] of all itemsets
      that are >= `min_support` and < than `max_len`
      (if `max_len` is not None).
      Each itemset in the 'itemsets' column is of type `frozenset`,
      which is a Python built-in type that behaves similarly to
      sets except that it is immutable
      (For more info, see
      https://docs.python.org/3.6/library/stdtypes.html#frozenset).

    Examples
    ----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/hmine/

    """

    fpc.valid_input_check(df)
    if min_support <= 0.0:
        raise ValueError(
            "`min_support` must be a positive "
            "number within the interval `(0, 1]`. "
            "Got %s." % min_support
        )
    # Calculate the minimum support based on the number of transactions (absolute support)
    minsup = math.ceil(min_support * len(df))

    is_sparse = False
    if hasattr(df, "sparse"):
        # DataFrame with SparseArray (pandas >= 0.24)
        if df.size == 0:
            itemsets = df.values
        else:
            itemsets = df.sparse.to_coo().tocsr()
            is_sparse = True
    else:
        # dense DataFrame
        itemsets = df.values
    if is_sparse:
        is_sparse
    single_items = np.array(df.columns)
    itemsets_shape = itemsets.shape[0]
    itemsets, single_items, single_items_support = itemset_optimisation(
        itemsets, single_items, minsup
    )
    numeric_single_items = np.arange(len(single_items))
    frequent_itemsets = {}
    for item in numeric_single_items:
        if single_items_support[item] >= minsup:
            supp = single_items_support[item] / itemsets_shape
            frequent_itemsets[frozenset([single_items[item]])] = supp
        if max_len == 1:
            continue
        # Recursive call to find frequent itemsets
        frequent_itemsets = hmine_driver(
            [item],
            itemsets,
            minsup,
            itemsets_shape,
            max_len,
            verbose,
            single_items,
            frequent_itemsets,
        )
    # Convert the dictionary to a DataFrame
    res_df = pd.DataFrame([frequent_itemsets.values(), frequent_itemsets.keys()]).T
    res_df.columns = ["support", "itemsets"]

    if not use_colnames:
        mapping = {item: idx for idx, item in enumerate(df.columns)}
        res_df["itemsets"] = res_df["itemsets"].apply(
            lambda x: frozenset([mapping[i] for i in x])
        )

    return res_df


def itemset_optimisation(
    itemsets: np.array,
    single_items: np.array,
    minsup: int,
) -> tuple:
    """
    Downward-closure property of H-Mine algorithm.
        Optimizes the itemsets matrix by removing items that do not
        meet the minimum support. (For more info, see
        https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/hmine/)

    Args:
        itemsets (np.array): matrix of bools or binary
        single_items (np.array): array of single items
        minsup (int): minimum absolute support

    Returns:
        itemsets (np.array): reduced itemsets matrix of bools or binary
        single_items (np.array): reduced array of single items
        single_items_support (np.array): reduced single items support
    """

    single_items_support = np.array(np.sum(itemsets, axis=0)).reshape(-1)
    items = np.nonzero(single_items_support >= minsup)[0]
    itemsets = itemsets[:, items]
    single_items = single_items[items]
    single_items_support = single_items_support[items]

    return itemsets, single_items, single_items_support


def hmine_driver(
    item: list,
    itemsets: np.array,
    minsup: int,
    itemsets_shape: int,
    max_len: int,
    verbose: int,
    single_items: np.array,
    frequent_itemsets: dict,
) -> dict:
    """
    Driver function for the hmine algorithm.
    Recursively generates frequent itemsets.
    Also works for sparse matrix.
    egg: item = [1] -> [1,2] -> [1,2,3] -> [1,2,4] -> [1,2,5]
    For more info, see
    https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/hmine/

    Args:
        item (list): list of items
        itemsets (np.array): matrix of bools or binary
        minsup (int): minimum absolute support
        itemsets_shape (int): number of transactions
        single_items (np.array): array of single items
        max_len (int): maximum length of frequent itemsets
        verbose (int): verbose mode
        frequent_itemsets (dict): dictionary of frequent itemsets

    Returns:
        frequent_itemsets(dict): dictionary of frequent itemsets
    """
    # Early stopping if the length of the item is greater than max_len
    if max_len and len(item) >= max_len:
        return frequent_itemsets
    projected_itemsets = create_projected_itemsets(item, itemsets)
    initial_supports = np.array(np.sum(projected_itemsets, axis=0)).reshape(-1)
    suffixes = np.nonzero(initial_supports >= minsup)[0]
    suffixes = suffixes[np.nonzero(suffixes > item[-1])[0]]

    if verbose:
        print(
            f"{len(suffixes)} itemset(s) from the suffixes on item(s) ({', '.join(single_items[item])})"
        )
    for suffix in suffixes:
        new_item = item.copy()
        new_item.append(suffix)
        supp = initial_supports[suffix] / itemsets_shape
        frequent_itemsets[frozenset(single_items[new_item])] = supp
        # Recursive call to find frequent itemsets
        frequent_itemsets = hmine_driver(
            new_item,
            projected_itemsets,
            minsup,
            itemsets_shape,
            max_len,
            verbose,
            single_items,
            frequent_itemsets,
        )
    return frequent_itemsets


def create_projected_itemsets(item: list, itemsets: np.array) -> np.array:
    """
    Creates the projected itemsets for the given item. (For more info, see
    https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/hmine/)

    Args:
        item (list): list of items
        itemsets (np.array): matrix of bools or binary

    Returns:
        projected_itemsets(np.array): projected itemsets for the given item
    """

    indices = np.nonzero(np.sum(itemsets[:, item], axis=1) == len(item))[0]
    projected_itemsets = itemsets[indices, :]
    return projected_itemsets



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for plotting decision regions of classifiers.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np


def heatmap(
    matrix,
    hide_spines=False,
    hide_ticks=False,
    figsize=None,
    cmap=None,
    colorbar=True,
    row_names=None,
    column_names=None,
    column_name_rotation=45,
    cell_values=True,
    cell_fmt=".2f",
    cell_font_size=None,
    text_color_threshold=None,
):
    """Plot a heatmap via matplotlib.

    Parameters
    -----------
    conf_mat : array-like, shape = [n_rows, n_columns]
        And arbitrary 2D array.

    hide_spines : bool (default: False)
        Hides axis spines if True.

    hide_ticks : bool (default: False)
        Hides axis ticks if True

    figsize : tuple (default: (2.5, 2.5))
        Height and width of the figure

    cmap : matplotlib colormap (default: `None`)
        Uses matplotlib.pyplot.cm.viridis if `None`

    colorbar : bool (default: True)
        Shows a colorbar if True

    row_names : array-like, shape = [n_rows] (default: None)
        List of row names to be used as y-axis tick labels.

    column_names : array-like, shape = [n_columns] (default: None)
        List of column names to be used as x-axis tick labels.

    column_name_rotation : int (default: 45)
        Number of degrees for rotating column x-tick labels.

    cell_values : bool (default: True)
        Plots cell values if True.

    cell_fmt : string (default: '.2f')
        Format specification for cell values (if `cell_values=True`)

    cell_font_size : int (default: None)
        Font size for cell values (if `cell_values=True`)

    text_color_threshold : float (default: None)
        Threshold for the black/white text threshold of the text
        annotation. Default (None) tried to infer a good threshold
        automatically using `np.max(normed_matrix) / 2`.

    Returns
    -----------
    fig, ax : matplotlib.pyplot subplot objects
        Figure and axis elements of the subplot.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/heatmap/

    """

    if row_names is not None and len(row_names) != matrix.shape[0]:
        raise AssertionError(
            f"len(row_names) (got {len(row_names)})"
            " should be equal to number of"
            " rows in the input "
            f" array (expect {matrix.shape[0]})."
        )

    if column_names is not None and len(column_names) != matrix.shape[1]:
        raise AssertionError(
            f"len(column_names)"
            " (got {len(column_names)})"
            " should be equal to number of"
            " columns in the"
            f" input array (expect {matrix.shape[1]})."
        )

    fig, ax = plt.subplots(figsize=figsize)
    ax.grid(False)

    if cmap is None:
        cmap = plt.cm.viridis

    if figsize is None:
        figsize = (len(matrix) * 1.25, len(matrix) * 1.25)

    matshow = ax.matshow(matrix, cmap=cmap)

    if colorbar:
        fig.colorbar(matshow)

    normed_matrix = matrix.astype("float") / matrix.max()
    if text_color_threshold is None:
        thres_value = np.max(normed_matrix) / 2
    else:
        thres_value = text_color_threshold

    if cell_values:
        for i in range(matrix.shape[0]):
            for j in range(matrix.shape[1]):
                cell_text = format(matrix[i, j], cell_fmt)

                ax.text(
                    x=j,
                    y=i,
                    size=cell_font_size,
                    s=cell_text,
                    va="center",
                    ha="center",
                    color="white" if normed_matrix[i, j] > thres_value else "black",
                )

    if row_names is not None:
        tick_marks = np.arange(len(row_names))
        plt.yticks(tick_marks, row_names)

    if column_names is not None:
        tick_marks = np.arange(len(column_names))

        if column_name_rotation:
            plt.xticks(
                tick_marks,
                column_names,
                rotation=column_name_rotation,
                ha="right",
                rotation_mode="anchor",
            )
        else:
            plt.xticks(tick_marks, column_names)

    if hide_spines:
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
    ax.yaxis.set_ticks_position("left")
    ax.xaxis.set_ticks_position("bottom")
    if hide_ticks:
        ax.axes.get_yaxis().set_ticks([])
        ax.axes.get_xaxis().set_ticks([])

    return fig, ax



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from packaging.version import Version
from scipy.sparse import issparse
from sklearn import __version__ as sklearn_version
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from mlxtend.preprocessing import DenseTransformer

iris = load_iris()
X, y = iris.data, iris.target


def test_dense_to_dense():
    todense = DenseTransformer(return_copy=False)
    np.testing.assert_array_equal(X, todense.transform(X))


def test_sparse_to_dense():
    todense = DenseTransformer()
    tfidf = TfidfTransformer()
    X_t = tfidf.fit_transform([[1, 2, 3]])
    assert issparse(X_t)
    X_dense = todense.transform(X_t)
    expect = np.array([[0.26726124, 0.53452248, 0.80178373]])
    assert np.allclose(X_dense, expect)


def test_pipeline():
    rf = RandomForestClassifier(n_estimators=10)
    param_grid = [{"randomforestclassifier__n_estimators": [1, 5, 10]}]
    pipe = make_pipeline(StandardScaler(), DenseTransformer(), rf)
    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(pipe, param_grid, cv=3, n_jobs=1, iid=False)
    else:
        grid = GridSearchCV(pipe, param_grid, cv=3, n_jobs=1)
    grid.fit(X, y)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


from mlxtend.data import (
    autompg_data,
    boston_housing_data,
    iris_data,
    mnist_data,
    three_blobs_data,
    wine_data,
)


def test_import_wine_data():
    X, y = wine_data()

    assert X.shape[0] == 178
    assert X.shape[1] == 13
    assert y.shape[0] == 178


def test_import_iris_data():
    X, y = iris_data()
    assert X.shape[0] == 150, X.shape
    assert X.shape[1] == 4
    assert y.shape[0] == 150


def test_import_autompg_data():
    X, y = autompg_data()
    assert X.shape[0] == 392
    assert X.shape[1] == 8
    assert y.shape[0] == 392


def test_import_mnist_data():
    X, y = mnist_data()
    assert X.shape[0] == 5000
    assert X.shape[1] == 784
    assert y.shape[0] == 5000


def test_import_three_blobs_data():
    X, y = three_blobs_data()
    assert X.shape[0] == 150, X.shape
    assert X.shape[1] == 2
    assert y.shape[0] == 150


def test_import_boston_housing():
    X, y = boston_housing_data()
    assert X.shape[0] == 506
    assert X.shape[1] == 13
    assert y.shape[0] == 506



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pytest

from mlxtend.preprocessing import one_hot


def test_default():
    y = np.array([0, 1, 2, 3, 4, 2])
    expect = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
        ],
        dtype="float",
    )
    out = one_hot(y)
    np.testing.assert_array_equal(expect, out)


def test_autoguessing():
    y = np.array([0, 4, 0, 4])
    expect = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
        ],
        dtype="float",
    )
    out = one_hot(y)
    np.testing.assert_array_equal(expect, out)


def test_list():
    y = [0, 1, 2, 3, 4, 2]
    expect = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
        ],
        dtype="float",
    )
    out = one_hot(y)
    np.testing.assert_array_equal(expect, out)


def test_multidim_list():
    y = [[0, 1, 2, 3, 4, 2]]
    with pytest.raises(AttributeError):
        one_hot(y)


def test_multidim_array():
    y = np.array([[0], [1], [2], [3], [4], [2]])
    with pytest.raises(AttributeError):
        one_hot(y)


def test_oneclass():
    np.testing.assert_array_equal(one_hot([0]), np.array([[0.0]], dtype="float"))


def test_list_morelabels():
    y = [0, 1]
    expect = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]], dtype="float")
    out = one_hot(y, num_labels=3)
    np.testing.assert_array_equal(expect, out)



from copy import deepcopy

import numpy as np
from sklearn.model_selection import cross_val_score


def _merge_lists(nested_list, high_level_indices=None):
    """
    merge elements of lists (of a nested_list) into one single tuple with elements
    sorted in ascending order.

    Parameters
    ----------
    nested_list: List
        a  list whose elements must be list as well.

    high_level_indices: list or tuple, default None
        a list or tuple that contains integers that are between 0 (inclusive) and
        the length of `nested_lst` (exclusive). If None, the merge of all
        lists nested in `nested_list` will be returned.

    Returns
    -------
    out: tuple
        a tuple, with elements sorted in ascending order, that is the merge of inner
        lists whose indices are provided in `high_level_indices`

    Example:
    nested_list = [[1],[2, 3],[4]]
    high_level_indices = [1, 2]
    >>> _merge_lists(nested_list, high_level_indices)
    (2, 3, 4) # merging [2, 3] and [4]
    """
    if high_level_indices is None:
        high_level_indices = list(range(len(nested_list)))

    out = []
    for idx in high_level_indices:
        out.extend(nested_list[idx])

    return tuple(sorted(out))


def _calc_score(
    selector, X, y, indices, groups=None, feature_groups=None, **fit_params
):
    """
    calculate the cross-validation score for feature data `X` and target variable
     `y`.

    Parameters
    ---------
    selector : objcet with attributes est_` (estimator), `cv` (number of folds
    in cross-validation), and `pre_dispatch`()

    X : numpy.ndarray
        A 2D array consisting of feature data, where each column corresponds to
        one feature, and each row corresponds to one instance (or observation)

    y : numpy.ndarray
        A 1D array consiting of tartget values

    indices : list or tuple
        A list or tuple of interger numbers. When `feature_groups` is not provided,
        i.e. None (default), the values in indices represent the column indices of
        X that should be consdered throughout the calculation of cross validation
        score. When `feature_groups` is not None, the indices represent the indices
        of the groups of features that should be considered through the calculation
        of cross-validation score.

    groups : array-like, with shape (n_samples,), optional
        Group labels for the samples used while splitting the dataset into
        train/test set. Passed to the fit method of the cross-validator.

    feature_groups : list or None (default: None)
        Optional argument for treating certain features as a group.
        This means, the features within a group are always selected together,
        never split.
        For example, `feature_groups=[[1], [2], [3, 4, 5]]`
        specifies 3 feature groups.e

    fit_params : dict of string -> object, optional
        Parameters to pass to to the fit method of classifier.

    Returns
    -------
    indices : List or tuple
        This is exactly the same as the input `indices`

    scores : array
        This is an array of cv scores, with length equal to the cv value.
    """
    if feature_groups is None:
        feature_groups = [[i] for i in range(X.shape[1])]

    IDX = _merge_lists(feature_groups, indices)
    if selector.cv:
        scores = cross_val_score(
            selector.est_,
            X[:, IDX],
            y,
            groups=groups,
            cv=selector.cv,
            scoring=selector.scorer,
            n_jobs=1,
            pre_dispatch=selector.pre_dispatch,
            fit_params=fit_params,
        )
    else:
        selector.est_.fit(X[:, IDX], y, **fit_params)
        scores = np.array([selector.scorer(selector.est_, X[:, IDX], y)])
    return indices, scores


def _preprocess(X):
    """
    Check if X is a DataFrame or not, and returns numpy ndarray and name of features.

    Parameters
    ----------
    X : DataFrame or numpy.ndarray
        A DataFrame or a 2D numpy.ndarray

    Returns
    -------
    X_ : numpy.ndarray
        A 2D array that is equivalanet to X.to_numpy()

    features_names : List
        A list consisting of name of features. When `X` is a DataFrame, it contains
        the name of columns. If it is a 2D array, it is None.
    """
    if X.ndim != 2:
        raise ValueError(f"The input X must be 2D array. Got {X.ndim}")

    if type(X).__name__ == "DataFrame":
        features_names = list(X.columns)
        X_ = X.to_numpy(copy=True)
    else:
        # it is numpy array
        features_names = None
        X_ = X.copy()

    return X_, features_names


def _get_featurenames(subsets_dict, feature_idx, feature_names, n_features):
    if feature_names is None:
        feature_names = [str(i) for i in range(n_features)]

    dict_keys = subsets_dict.keys()
    for key in dict_keys:
        subsets_dict[key]["feature_names"] = tuple(
            feature_names[idx] for idx in subsets_dict[key]["feature_idx"]
        )

    if feature_idx is None:
        feature_idx_names = None
    else:
        feature_idx_names = tuple(feature_names[idx] for idx in feature_idx)

    return subsets_dict, feature_idx_names



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend.evaluate import bootstrap
from mlxtend.utils import assert_raises


def test_defaults():
    rng = np.random.RandomState(123)
    x = rng.normal(loc=5.0, size=100)
    original, std_err, ci_bounds = bootstrap(x, func=np.mean, seed=123)
    assert round(original, 2) == 5.03
    assert round(std_err, 2) == 0.11
    assert round(ci_bounds[0], 2) == 4.80
    assert round(ci_bounds[1], 2) == 5.26


def test_fail_ci():
    msg = "ci must be in range (0, 1)"
    assert_raises(
        AttributeError, msg, bootstrap, np.array([1, 2, 3]), np.mean, 1000, 95
    )


def test_fail_func():
    def f(x):
        return np.array([0, 1])

    msg = "ci must be in range (0, 1)"
    assert_raises(AttributeError, msg, bootstrap, np.array([1, 2, 3]), f, 1000, 95)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pytest

from mlxtend.feature_extraction.base import _BaseFeatureExtractor

X, y = np.array([[1, 2], [4, 5], [3, 9]]), np.array([1, 2, 3])
base = _BaseFeatureExtractor()


def test_X_array_pass():
    base._check_arrays(X=X)


def test_X_y_array_pass():
    base._check_arrays(X=X, y=y)


def test_1D_X():
    with pytest.raises(ValueError):
        base._check_arrays(X=X[1])


def test_X_int_y():
    with pytest.raises(ValueError):
        base._check_arrays(X=X, y=y[1])


def test_X_short_y():
    print(y[1:].shape)
    print(X.shape)
    with pytest.raises(ValueError):
        base._check_arrays(X=X, y=y[1:])



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import random

import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from packaging.version import Version
from scipy import sparse
from sklearn import __version__ as sklearn_version
from sklearn import exceptions
from sklearn.base import clone
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression, PassiveAggressiveClassifier
from sklearn.model_selection import GridSearchCV, cross_val_score, train_test_split
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC

from mlxtend.classifier import StackingClassifier
from mlxtend.data import iris_data
from mlxtend.externals.estimator_checks import NotFittedError
from mlxtend.utils import assert_raises

X, y = iris_data()
X = X[:, 1:3]


y2 = np.c_[y, y]


def test_StackingClassifier():
    np.random.seed(123)
    meta = LogisticRegression(
        solver="liblinear",
        multi_class="ovr",
    )
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(classifiers=[clf1, clf2], meta_classifier=meta)

    scores = cross_val_score(sclf, X, y, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)

    if Version(sklearn_version) < Version("0.20"):
        assert scores_mean == 0.95, scores_mean
    else:
        assert scores_mean == 0.95, scores_mean


def test_fit_base_estimators_false():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()

    clf1.fit(X, y)
    clf2.fit(X, y)

    sclf = StackingClassifier(
        classifiers=[clf1, clf2], meta_classifier=meta, fit_base_estimators=False
    )

    sclf.fit(X, y)
    assert round(sclf.score(X, y), 2) == 0.98


def test_use_clones():
    np.random.seed(123)
    X, y = iris_data()

    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    StackingClassifier(
        classifiers=[clf1, clf2], use_clones=True, meta_classifier=meta
    ).fit(X, y)

    assert_raises(
        exceptions.NotFittedError,
        "This RandomForestClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this estimator.",
        clf1.predict,
        X,
    )

    StackingClassifier(
        classifiers=[clf1, clf2],
        use_probas=True,
        use_clones=False,
        meta_classifier=meta,
    ).fit(X, y)

    clf1.predict(X)


def test_sample_weight():
    # Make sure that:
    #    prediction with weight
    # != prediction with no weight
    # == prediction with weight ones
    random.seed(87)
    w = np.array([random.random() for _ in range(len(y))])

    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(classifiers=[clf1, clf2], meta_classifier=meta)
    prob1 = sclf.fit(X, y, sample_weight=w).predict_proba(X)

    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(classifiers=[clf1, clf2], meta_classifier=meta)
    prob2 = sclf.fit(X, y, sample_weight=None).predict_proba(X)

    maxdiff = np.max(np.abs(prob1 - prob2))
    assert maxdiff > 1e-3, "max diff is %.4f" % maxdiff

    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(classifiers=[clf1, clf2], meta_classifier=meta)
    prob3 = sclf.fit(X, y, sample_weight=np.ones(len(y))).predict_proba(X)

    maxdiff = np.max(np.abs(prob2 - prob3))
    assert maxdiff < 1e-3, "max diff is %.4f" % maxdiff


def test_weight_unsupported():
    # Error since KNN does not support sample_weight
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    clf3 = KNeighborsClassifier()
    sclf = StackingClassifier(classifiers=[clf1, clf2, clf3], meta_classifier=meta)
    random.seed(87)
    w = np.array([random.random() for _ in range(len(y))])

    with pytest.raises(TypeError):
        sclf.fit(X, y, sample_weight=w)


def test_weight_unsupported_no_weight():
    # This is okay since we do not pass sample weight
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    clf3 = KNeighborsClassifier()
    sclf = StackingClassifier(classifiers=[clf1, clf2, clf3], meta_classifier=meta)
    sclf.fit(X, y)


def test_StackingClassifier_proba_avg_1():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr", random_state=1)
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, X, y, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.93, scores_mean


def test_StackingClassifier_proba_concat_1():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2],
        use_probas=True,
        average_probas=False,
        meta_classifier=meta,
    )

    scores = cross_val_score(sclf, X, y, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.93, scores_mean


def test_StackingClassifier_avg_vs_concat():
    np.random.seed(123)
    lr1 = LogisticRegression(solver="liblinear", multi_class="ovr")
    sclf1 = StackingClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        average_probas=True,
        meta_classifier=lr1,
    )

    sclf1.fit(X, y)
    r1 = sclf1.predict_meta_features(X[:2])
    assert r1.shape == (2, 3)
    assert_almost_equal(np.sum(r1[0]), 1.0, decimal=6)
    assert_almost_equal(np.sum(r1[1]), 1.0, decimal=6)

    sclf2 = StackingClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        average_probas=False,
        meta_classifier=lr1,
    )

    sclf2.fit(X, y)
    r2 = sclf2.predict_meta_features(X[:2])
    assert r2.shape == (2, 6)
    assert_almost_equal(np.sum(r2[0]), 2.0, decimal=6)
    assert_almost_equal(np.sum(r2[1]), 2.0, decimal=6)
    np.array_equal(r2[0][:3], r2[0][3:])


def test_StackingClassifier_drop_proba_col():
    np.random.seed(123)
    lr1 = LogisticRegression(solver="liblinear", multi_class="ovr")
    sclf1 = StackingClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col=None,
        meta_classifier=lr1,
    )

    sclf1.fit(X, y)
    r1 = sclf1.predict_meta_features(X[:2])
    assert r1.shape == (2, 6)

    sclf2 = StackingClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col="last",
        meta_classifier=lr1,
    )

    sclf2.fit(X, y)
    r2 = sclf2.predict_meta_features(X[:2])
    assert r2.shape == (2, 4), r2.shape

    sclf4 = StackingClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col="first",
        meta_classifier=lr1,
    )

    sclf4.fit(X, y)
    r4 = sclf4.predict_meta_features(X[:2])
    assert r4.shape == (2, 4), r4.shape

    sclf3 = StackingClassifier(
        classifiers=[lr1, lr1],
        use_probas=True,
        drop_proba_col="last",
        meta_classifier=lr1,
    )

    sclf3.fit(X[0:100], y[0:100])  # only 2 classes
    r3 = sclf3.predict_meta_features(X[:2])
    assert r3.shape == (2, 2), r3.shape


def test_multivariate_class():
    np.random.seed(123)
    meta = KNeighborsClassifier()
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = KNeighborsClassifier()
    sclf = StackingClassifier(classifiers=[clf1, clf2], meta_classifier=meta)
    y_pred = sclf.fit(X, y2).predict(X)
    ca = 0.973
    assert round((y_pred == y2).mean(), 3) == ca


def test_gridsearch():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(classifiers=[clf1, clf2], meta_classifier=meta)

    params = {
        "meta_classifier__C": [1.0, 100.0],
        "randomforestclassifier__n_estimators": [20, 200],
    }

    grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5)
    X, y = iris_data()
    grid.fit(X, y)

    mean_scores = [round(s, 2) for s in grid.cv_results_["mean_test_score"]]

    assert mean_scores == [0.95, 0.97, 0.96, 0.96], mean_scores


def test_gridsearch_enumerate_names():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(classifiers=[clf1, clf1, clf2], meta_classifier=meta)

    params = {
        "meta_classifier__C": [1.0, 100.0],
        "randomforestclassifier-1__n_estimators": [5, 10],
        "randomforestclassifier-2__n_estimators": [5, 20],
        "use_probas": [True, False],
    }

    grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5)
    X, y = iris_data()
    grid = grid.fit(X, y)


def test_use_probas():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, X, y, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.93, scores_mean


def test_not_fitted():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    X, _ = iris_data()

    assert_raises(
        NotFittedError,
        "This StackingClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this method.",
        sclf.predict,
        X,
    )

    assert_raises(
        NotFittedError,
        "This StackingClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this method.",
        sclf.predict_proba,
        X,
    )

    assert_raises(
        NotFittedError,
        "This StackingClassifier instance is not fitted yet."
        " Call 'fit' with appropriate arguments"
        " before using this method.",
        sclf.predict_meta_features,
        X,
    )


def test_verbose():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta, verbose=3
    )
    X, y = iris_data()
    sclf.fit(X, y)


def test_use_features_in_secondary_predict():
    np.random.seed(123)
    X, y = iris_data()
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_features_in_secondary=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, X, y, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.95, scores_mean


def test_use_features_in_secondary_predict_proba():
    np.random.seed(123)
    X, y = iris_data()
    meta = LogisticRegression(solver="liblinear", multi_class="ovr", random_state=1)
    clf1 = RandomForestClassifier(n_estimators=10, random_state=1)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_features_in_secondary=True, meta_classifier=meta
    )

    sclf.fit(X, y)
    idx = [0, 1, 2]
    y_pred = sclf.predict_proba(X[idx])[:, 0]
    expect = np.array([0.916, 0.828, 0.889])
    np.testing.assert_almost_equal(y_pred, expect, 3)


def test_use_features_in_secondary_sparse_input_predict():
    np.random.seed(123)
    X, y = iris_data()
    meta = LogisticRegression(solver="liblinear", multi_class="ovr", random_state=1)
    clf1 = RandomForestClassifier(n_estimators=10, random_state=1)
    sclf = StackingClassifier(
        classifiers=[clf1], use_features_in_secondary=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, sparse.csr_matrix(X), y, cv=5, scoring="accuracy")
    scores_mean = round(scores.mean(), 2)
    assert scores_mean == 0.97, scores_mean


def test_use_features_in_secondary_sparse_input_predict_proba():
    np.random.seed(123)
    meta = LogisticRegression(solver="liblinear", multi_class="ovr")
    clf1 = RandomForestClassifier(n_estimators=10)
    sclf = StackingClassifier(
        classifiers=[clf1], use_features_in_secondary=True, meta_classifier=meta
    )

    sclf.fit(sparse.csr_matrix(X), y)
    idx = [0, 1, 2]
    y_pred = sclf.predict_proba(sparse.csr_matrix(X[idx]))[:, 0]
    expect = np.array([0.910, 0.829, 0.882])
    np.testing.assert_almost_equal(y_pred, expect, 3)


def test_get_params():
    np.random.seed(123)
    clf1 = KNeighborsClassifier(n_neighbors=1)
    clf2 = RandomForestClassifier(n_estimators=10)
    clf3 = GaussianNB()
    lr = LogisticRegression(solver="liblinear", multi_class="ovr")
    sclf = StackingClassifier(classifiers=[clf1, clf2, clf3], meta_classifier=lr)

    got = sorted(list({s.split("__")[0] for s in sclf.get_params().keys()}))
    expect = [
        "average_probas",
        "classifiers",
        "drop_proba_col",
        "fit_base_estimators",
        "gaussiannb",
        "kneighborsclassifier",
        "meta_classifier",
        "randomforestclassifier",
        "store_train_meta_features",
        "use_clones",
        "use_features_in_secondary",
        "use_probas",
        "verbose",
    ]
    assert got == expect, got


def test_classifier_gridsearch():
    np.random.seed(123)
    clf1 = KNeighborsClassifier(n_neighbors=1)
    clf2 = RandomForestClassifier(n_estimators=10)
    clf3 = GaussianNB()
    lr = LogisticRegression(solver="liblinear", multi_class="ovr")
    sclf = StackingClassifier(classifiers=[clf1, clf2, clf3], meta_classifier=lr)

    params = {"classifiers": [[clf1, clf1, clf1], [clf2, clf3]]}

    grid = GridSearchCV(estimator=sclf, param_grid=params, cv=5, refit=True)
    grid.fit(X, y)

    assert len(grid.best_params_["classifiers"]) == 2


def test_train_meta_features_():
    np.random.seed(123)
    knn = KNeighborsClassifier()
    lr = LogisticRegression(solver="liblinear", multi_class="ovr")
    gnb = GaussianNB()
    stclf = StackingClassifier(
        classifiers=[knn, gnb], meta_classifier=lr, store_train_meta_features=True
    )
    X_train, _, y_train, _ = train_test_split(X, y, test_size=0.3)
    stclf.fit(X_train, y_train)
    train_meta_features = stclf.train_meta_features_
    assert train_meta_features.shape == (X_train.shape[0], 2)


def test_predict_meta_features():
    knn = KNeighborsClassifier()
    lr = LogisticRegression(solver="liblinear", multi_class="ovr", random_state=1)
    gnb = GaussianNB()
    X_train, X_test, y_train, _ = train_test_split(X, y, test_size=0.3)

    #  test default (class labels)
    stclf = StackingClassifier(
        classifiers=[knn, gnb], meta_classifier=lr, store_train_meta_features=True
    )
    stclf.fit(X_train, y_train)
    test_meta_features = stclf.predict(X_test)
    assert test_meta_features.shape == (X_test.shape[0],)


def test_clone():
    np.random.seed(1)
    knn = KNeighborsClassifier()
    lr = LogisticRegression(solver="liblinear", multi_class="ovr")
    gnb = GaussianNB()
    stclf = StackingClassifier(
        classifiers=[knn, gnb], meta_classifier=lr, store_train_meta_features=True
    )
    clone(stclf)


def test_decision_function():
    np.random.seed(123)

    # PassiveAggressiveClassifier has no predict_proba
    meta = PassiveAggressiveClassifier(max_iter=1000, tol=0.001, random_state=42)
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    # binarize target
    y2 = y > 1
    scores = cross_val_score(sclf, X, y2, cv=5, scoring="roc_auc")
    scores_mean = round(scores.mean(), 2)

    if Version(sklearn_version) < Version("0.21"):
        assert scores_mean == 0.96, scores_mean
    else:
        assert scores_mean == 0.93, scores_mean

    # another test
    meta = SVC(decision_function_shape="ovo")

    sclf = StackingClassifier(
        classifiers=[clf1, clf2], use_probas=True, meta_classifier=meta
    )

    scores = cross_val_score(sclf, X, y2, cv=5, scoring="roc_auc")
    scores_mean = round(scores.mean(), 2)

    if Version(sklearn_version) < Version("0.22"):
        assert scores_mean == 0.95, scores_mean
    else:
        assert scores_mean == 0.94, scores_mean


def test_drop_col_unsupported():
    np.random.seed(123)
    meta = LogisticRegression()
    clf1 = RandomForestClassifier(n_estimators=10)
    clf2 = GaussianNB()
    clf3 = KNeighborsClassifier()

    with pytest.raises(ValueError):
        StackingClassifier(
            classifiers=[clf1, clf2, clf3],
            meta_classifier=meta,
            drop_proba_col="invalid value",
        )



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from mlxtend.utils import assert_raises


def test_without_message():
    def my_func():
        raise AttributeError

    assert_raises(AttributeError, func=my_func, message=None)


def test_with_message():
    def my_func():
        raise AttributeError("Failed")

    assert_raises(AttributeError, func=my_func, message="Failed")



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from mlxtend.data import iris_data
from mlxtend.plotting import ecdf


def test_threshold():
    X, y = iris_data()
    ax, threshold, count = ecdf(x=X[:, 0], x_label="sepal length (cm)", percentile=0.8)
    assert threshold == 6.5
    assert count == 120



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np
import pytest

from mlxtend.plotting import heatmap

plt.switch_backend("agg")


def test_defaults():
    heatmap(np.random.random((10, 5)))


def test_wrong_column_name_number():
    with pytest.raises(AssertionError) as excinfo:
        heatmap(np.random.random((10, 5)), column_names=["a", "b", "c"])
        assert excinfo.value.message == (
            "len(column_names) (got 3)"
            " should be equal to number of"
            " rows in the input "
            " array (expect 5)."
        )


def test_wrong_row_name_number():
    with pytest.raises(AssertionError) as excinfo:
        heatmap(np.random.random((10, 5)), row_names=["a", "b", "c"])
        assert excinfo.value.message == (
            "len(column_names) (got 3)"
            " should be equal to number of"
            " rows in the input "
            " array (expect 10)."
        )



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend.evaluate import permutation_test
from mlxtend.utils import assert_raises

treatment = [689, 656, 668, 660, 679, 663, 664, 647]
control = [657, 623, 652, 654, 658, 660, 670, 620]


def test_one_sided_x_greater_y():
    p = permutation_test(treatment, control, func=lambda x, y: np.mean(x) - np.mean(y))
    assert round(p, 4) == 0.0301, p

    p = permutation_test(treatment, control, func="x_mean > y_mean")
    assert round(p, 4) == 0.0301, p


def test_one_sided_y_greater_x():
    p = permutation_test(treatment, control, func=lambda x, y: np.mean(y) - np.mean(x))
    assert round(p, 3) == 0.973, p

    p = permutation_test(treatment, control, func="x_mean < y_mean")
    assert round(p, 3) == 0.973, p


def test_two_sided():
    p = permutation_test(
        treatment, control, func=lambda x, y: np.abs(np.mean(x) - np.mean(y))
    )
    assert round(p, 3) == 0.060, p

    p = permutation_test(treatment, control, func="x_mean != y_mean")
    assert round(p, 3) == 0.060, p


def test_default():
    p = permutation_test(treatment, control)
    assert round(p, 3) == 0.060, p


def test_approximateone_sided_x_greater_y():
    p = permutation_test(
        treatment,
        control,
        func=lambda x, y: np.mean(x) - np.mean(y),
        method="approximate",
        num_rounds=5000,
        seed=123,
    )
    assert round(p, 3) == 0.031, p


def test_invalid_method():
    msg = 'method must be "approximate" or "exact", got na'
    assert_raises(
        AttributeError,
        msg,
        permutation_test,
        [1, 2, 3],
        [3, 4, 5],
        lambda x, y: np.mean(x) - np.mean(y),
        method="na",
    )


def test_invalid_func():
    msg = (
        "Provide a custom function lambda x,y: ... "
        'or a string in ("x_mean != y_mean", '
        '"x_mean > y_mean", "x_mean < y_mean")'
    )
    assert_raises(AttributeError, msg, permutation_test, [1, 2, 3], [3, 4, 5], "myfunc")


def test_paired_runs_approximate():
    a = [3.67, 1.72, 3.46, 2.60, 2.03, 2.10, 3.01]
    b = [2.11, 1.79, 2.71, 1.89, 1.69, 1.71, 2.01]
    p = permutation_test(
        a, b, paired=True, method="approximate", seed=0, num_rounds=100000
    )
    assert round(p, 3) == 0.031


def test_paired_runs_exact():
    a = [3.67, 1.72, 3.46, 2.60, 2.03, 2.10, 3.01]
    b = [2.11, 1.79, 2.71, 1.89, 1.69, 1.71, 2.01]
    p = permutation_test(a, b, paired=True, method="exact")
    assert round(p, 3) == 0.031


def test_paired_invalid_lengths():
    msg = "x and y must have the same" " length if `paired=True`"
    assert_raises(ValueError, msg, permutation_test, [1, 2, 3], [3, 4], paired=True)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


import pandas as pd
from sklearn.datasets import load_iris
from sklearn.neighbors import KNeighborsClassifier

from mlxtend.feature_selection import SequentialFeatureSelector as SFS
from mlxtend.utils import assert_raises

iris = load_iris()
X_iris = iris.data
y_iris = iris.target


def test_run_forward():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=3,
        forward=True,
        floating=False,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_floating_forward():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=3,
        forward=True,
        floating=True,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_backward():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=3,
        forward=False,
        floating=False,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_floating_backward():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=3,
        forward=False,
        floating=True,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_forward_with_range():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=(3, 4),
        forward=True,
        floating=False,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_floating_forward_with_range():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=(3, 4),
        forward=True,
        floating=True,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_backward_with_range():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=(3, 4),
        forward=False,
        floating=False,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_run_floating_backward_with_range():
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=(3, 4),
        forward=False,
        floating=True,
        fixed_features=(0, 1),
        verbose=0,
    )
    sfs.fit(X_iris, y_iris)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_pandas():
    X_df = pd.DataFrame(
        X_iris, columns=["sepal length", "sepal width", "petal width", "petal width"]
    )
    knn = KNeighborsClassifier()
    sfs = SFS(
        estimator=knn,
        k_features=3,
        forward=True,
        floating=False,
        fixed_features=("sepal length", "sepal width"),
        verbose=0,
    )
    sfs.fit(X_df, y_iris)
    print(sfs.subsets_)
    for k in sfs.subsets_:
        assert 0 in sfs.subsets_[k]["feature_idx"]
        assert 1 in sfs.subsets_[k]["feature_idx"]


def test_wrong_feature_number():
    knn = KNeighborsClassifier()
    expect = "k_features must be between len(fixed_features) and X.shape[1]."
    sfs = SFS(knn, k_features=1, fixed_features=(1, 3))
    assert_raises(AttributeError, expect, sfs.fit, X=X_iris, y=y_iris)


def test_wrong_feature_range():
    knn = KNeighborsClassifier()
    expect = (
        "k_features tuple min value must be between len(fixed_features) and X.shape[1]."
    )
    sfs = SFS(knn, k_features=(1, 3), fixed_features=(1, 3))
    assert_raises(AttributeError, expect, sfs.fit, X=X_iris, y=y_iris)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for collecting file-group names from local directories.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import os
import re

from . import find_files


def find_filegroups(
    paths,
    substring="",
    extensions=None,
    validity_check=True,
    ignore_invisible=True,
    rstrip="",
    ignore_substring=None,
):
    """Find and collect files from different directories in a python dictionary.

    Parameters
    ----------
    paths : `list`
        Paths of the directories to be searched. Dictionary keys are build from
        the first directory.
    substring : `str` (default: '')
        Substring that all files have to contain to be considered.
    extensions : `list` (default: None)
        `None` or `list` of allowed file extensions for each path.
        If provided, the number of extensions must match the number of `paths`.
    validity_check : `bool` (default: None)
        If `True`, checks if all dictionary values
        have the same number of file paths. Prints
        a warning and returns an empty dictionary if the validity check failed.
    ignore_invisible : `bool` (default: True)
        If `True`, ignores invisible files
        (i.e., files starting with a period).
    rstrip : `str` (default: '')
        If provided, strips characters from right side of the file
        base names after splitting the extension.
        Useful to trim different filenames to a common stem.
        E.g,. "abc_d.txt" and "abc_d_.csv" would share
        the stem "abc_d" if rstrip is set to "_".
    ignore_substring : `str` (default: None)
        Ignores files that contain the specified substring.

    Returns
    ----------
    groups : `dict`
        Dictionary of files paths. Keys are the file names
        found in the first directory listed
        in `paths` (without file extension).

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/file_io/find_filegroups/

    """
    n = len(paths)

    # must have same number of paths and extensions
    assert len(paths) >= 2
    if extensions:
        assert len(extensions) == n
    else:
        extensions = ["" for i in range(n)]

    base = find_files(
        path=paths[0],
        substring=substring,
        check_ext=extensions[0],
        ignore_invisible=ignore_invisible,
        ignore_substring=ignore_substring,
    )
    rest = [
        find_files(
            path=paths[i],
            substring=substring,
            check_ext=extensions[i],
            ignore_invisible=ignore_invisible,
            ignore_substring=ignore_substring,
        )
        for i in range(1, n)
    ]

    groups = {}
    for f in base:
        basename = os.path.splitext(os.path.basename(f))[0]
        basename = re.sub(r"\%s$" % rstrip, "", basename)
        groups[basename] = [f]

    # groups = {os.path.splitext(os.path.basename(f))[0].rstrip(rstrip):[f]
    #           for f in base}

    for idx, r in enumerate(rest):
        for f in r:
            basename, ext = os.path.splitext(os.path.basename(f))
            basename = re.sub(r"\%s$" % rstrip, "", basename)
            try:
                if extensions[idx + 1] == "" or ext == extensions[idx + 1]:
                    groups[basename].append(f)
            except KeyError:
                pass

    if validity_check:
        lens = [len(groups[k]) for k in groups.keys()]
        if len(set(lens)) > 1:
            raise ValueError(
                "Warning, some keys have more/less values than"
                " others. Set validity_check=False"
                " to ignore this warning."
            )

    return groups



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Feature Importance Estimation Through Permutation
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np


def feature_importance_permutation(
    X, y, predict_method, metric, num_rounds=1, feature_groups=None, seed=None
):
    """Feature importance imputation via permutation importance

    Parameters
    ----------

    X : NumPy array, shape = [n_samples, n_features]
        Dataset, where n_samples is the number of samples and
        n_features is the number of features.

    y : NumPy array, shape = [n_samples]
        Target values.

    predict_method : prediction function
        A callable function that predicts the target values
        from X.

    metric : str, callable
        The metric for evaluating the feature importance through
        permutation. By default, the strings 'accuracy' is
        recommended for classifiers and the string 'r2' is
        recommended for regressors. Optionally, a custom
        scoring function (e.g., `metric=scoring_func`) that
        accepts two arguments, y_true and y_pred, which have
        similar shape to the `y` array.

    num_rounds : int (default=1)
        Number of rounds the feature columns are permuted to
        compute the permutation importance.

    feature_groups : list or None (default=None)
        Optional argument for treating certain features as a group.
        For example `[1, 2, [3, 4, 5]]`, which can be useful for
        interpretability, for example, if features 3, 4, 5 are one-hot
        encoded features.

    seed : int or None (default=None)
        Random seed for permuting the feature columns.

    Returns
    ---------

    mean_importance_vals, all_importance_vals : NumPy arrays.
      The first array, mean_importance_vals has shape [n_features, ] and
      contains the importance values for all features.
      The shape of the second array is [n_features, num_rounds] and contains
      the feature importance for each repetition. If num_rounds=1,
      it contains the same values as the first array, mean_importance_vals.

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/feature_importance_permutation/

    """

    if not isinstance(num_rounds, int):
        raise ValueError("num_rounds must be an integer.")
    if num_rounds < 1:
        raise ValueError("num_rounds must be greater than 1.")

    if not (metric in ("r2", "accuracy") or hasattr(metric, "__call__")):
        raise ValueError(
            'metric must be either "r2", "accuracy", '
            "or a function with signature func(y_true, y_pred)."
        )

    if metric == "r2":

        def score_func(y_true, y_pred):
            sum_of_squares = np.sum(np.square(y_true - y_pred))
            res_sum_of_squares = np.sum(np.square(y_true - y_true.mean()))
            r2_score = 1.0 - (sum_of_squares / res_sum_of_squares)
            return r2_score

    elif metric == "accuracy":

        def score_func(y_true, y_pred):
            return np.mean(y_true == y_pred)

    else:
        score_func = metric

    rng = np.random.RandomState(seed)

    baseline = score_func(y, predict_method(X))

    if feature_groups is None:
        mean_importance_vals = np.zeros(X.shape[1])
        all_importance_vals = np.zeros((X.shape[1], num_rounds))

        for round_idx in range(num_rounds):
            for col_idx in range(X.shape[1]):
                save_col = X[:, col_idx].copy()
                rng.shuffle(X[:, col_idx])
                new_score = score_func(y, predict_method(X))
                X[:, col_idx] = save_col
                importance = baseline - new_score
                mean_importance_vals[col_idx] += importance
                all_importance_vals[col_idx, round_idx] = importance
        mean_importance_vals /= num_rounds

    else:
        mean_importance_vals = np.zeros(len(feature_groups))
        all_importance_vals = np.zeros((len(feature_groups), num_rounds))
        for round_idx in range(num_rounds):
            for col_idx, feat in enumerate(feature_groups):
                save_col = X[:, feat].copy()

                if save_col.ndim > 1:
                    columns = save_col.shape[1]
                    for i in range(columns):
                        rng.shuffle(X[:, i])
                else:
                    rng.shuffle(X[:, feat])

                new_score = score_func(y, predict_method(X))
                X[:, feat] = save_col
                importance = baseline - new_score
                mean_importance_vals[col_idx] += importance
                all_importance_vals[col_idx, round_idx] = importance
        mean_importance_vals /= num_rounds

    return mean_importance_vals, all_importance_vals



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for loading the open-source Wine dataset.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import os

import numpy as np

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data", "wine.csv")


def wine_data():
    """Wine dataset.

    Source : https://archive.ics.uci.edu/ml/datasets/Wine
    Number of samples : 178
    Class labels : {0, 1, 2}, distribution: [59, 71, 48]

    Dataset Attributes:

        - 1) Alcohol
        - 2) Malic acid
        - 3) Ash
        - 4) Alcalinity of ash
        - 5) Magnesium
        - 6) Total phenols
        - 7) Flavanoids
        - 8) Nonflavanoid phenols
        - 9) Proanthocyanins
        - 10) Color intensity
        - 11) Hue
        - 12) OD280/OD315 of diluted wines
        - 13) Proline

    Returns
    --------
    X, y : [n_samples, n_features], [n_class_labels]
        X is the feature matrix with 178 wine samples as rows
        and 13 feature columns.
        y is a 1-dimensional array of the 3 class labels 0, 1, 2

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/data/wine_data

    """
    tmp = np.loadtxt(DATA_PATH, delimiter=",")
    X, y = tmp[:, :-1], tmp[:, -1]
    y = y.astype(int)
    return X, y



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np

from mlxtend.plotting import checkerboard_plot

plt.switch_backend("agg")


def test_runs():
    ary = np.random.random((6, 4))
    checkerboard_plot(
        ary,
        col_labels=["abc", "def", "ghi", "jkl"],
        row_labels=["sample %d" % i for i in range(1, 6)],
        cell_colors=["skyblue", "whitesmoke"],
        font_colors=["black", "black"],
        figsize=(5, 5),
    )
